from dataclasses import dataclass, field
from ..core.model import DataModel

from typing import Literal, Optional

from .component_types import *
from ..models import EmojiModel
from .component_types import ComponentTypes


@dataclass
class ActionRowPart(DataModel, ContainerChild):
    """Represents a container of interactable components."""
    
    components: list[ActionRowChild] = field(default_factory=list)
    """Up to 5 interactive button components or a single select component."""

    type: int = field(init=False, default=ComponentTypes.ACTION_ROW)
    """Component type. Always `ComponentTypes.ACTION_ROW` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

class ButtonStyles:
    """Represents button styles for a Button component."""

    PRIMARY = 1
    """The most important or recommended action in a group of options. (Blurple)"""

    SECONDARY = 2
    """Alternative or supporting actions. (Gray)"""

    SUCCESS = 3
    """Positive confirmation or completion actions. (Green)"""

    DANGER = 4
    """An action with irreversible consequences. (Red)"""

    LINK = 5
    """Navigates to a URL. (Gray + window)"""

@dataclass
class Button(DataModel, ActionRowChild, SectionAccessoryChild):
    """Represents the Button component."""

    style: int = None
    """A button style. See [`ButtonStyles`][scurrypy.parts.components.ButtonStyles]."""

    custom_id: str = None
    """ID for the button. Do not supply for `LINK` style buttons."""

    label: Optional[str] = None
    """Text that appears on the button."""

    emoji: Optional[EmojiModel] = None
    """Emoji icon for the button."""

    url: Optional[str] = None
    """URL for link-style buttons."""

    disabled: Optional[bool] = False
    """Whether the button is disabled. Defaults to `False`."""

    link: Optional[str] = None
    """Hyperlink for button. For `LINK` style only."""

    type: int = field(init=False, default=ComponentTypes.BUTTON)
    """Component type. Always `ComponentTypes.BUTTON` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

@dataclass
class SelectOption(DataModel):
    """Represents the Select Option component."""

    label: str = None
    """User-facing name of the option."""

    value: str = None
    """Developer-defined value of the option."""

    description: Optional[str] = None
    """Additional description of the option."""

    emoji: Optional[EmojiModel] = None
    """Partial emoji object."""

    default: Optional[bool] = False
    """Whether this option is selected by default."""

@dataclass
class StringSelect(DataModel, ActionRowChild, LabelChild):
    """Represents the String Select component."""

    custom_id: str = None
    """ID for the select menu."""

    options: list[SelectOption] = field(default_factory=list)
    """Specified choices in a select menu. See [`SelectOption`][scurrypy.parts.components.SelectOption]."""

    placeholder: Optional[str] = None
    """Placeholder text if nothing is selected or default."""

    min_values: Optional[int] = 1
    """Minimum number of items that must be chosen."""

    max_values: Optional[int] = 1
    """Maximum number of items that can be chosen."""

    required: Optional[bool] = False
    """Whether the string select is required to answer in a modal. Defaults to `False`."""

    disabled: Optional[bool] = False # does not work on Modals!
    """Whether select menu is disabled in a message. Defaults to `False`."""

    type: int = field(init=False, default=ComponentTypes.STRING_SELECT)
    """Component type. Always `ComponentTypes.STRING_SELECT` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

class TextInputStyles:
    """Represents the types of Text Inputs."""

    SHORT = 1
    """One line text input."""

    PARAGRAPH = 2
    """Multi-line text input."""

@dataclass
class TextInput(DataModel, LabelChild):
    """Represents the Text Input component."""

    custom_id: str = None
    """ID for the input."""

    style: TextInputStyles = TextInputStyles.SHORT
    """Text input style. See [`TextInputStyles`][scurrypy.parts.components.TextInputStyles]."""

    min_length: Optional[int] = None
    """Minimum input length for a text input."""

    max_length: Optional[int] = None
    """Maximum input length for a text input."""

    required: Optional[bool] = True
    """Whether this component is required to be filled. Defaults to True."""

    value: Optional[str] = None
    """Pre-filled value for this component."""

    placeholder: Optional[str] = None
    """Custom placeholder text if the input is empty."""

    type: int = field(init=False, default=ComponentTypes.TEXT_INPUT)
    """Component type. Always `ComponentTypes.TEXT_INPUT` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

@dataclass
class DefaultValue(DataModel):
    """Represents the Default Value for Select components."""

    id: int = None
    """ID of role, user, or channel."""

    type: Literal["role", "user", "channel"] = None
    """Type of value that `id` represents."""

@dataclass
class SelectMenu(DataModel):
    """Represents common fields for Discord's select menus."""

    custom_id: str = None
    """ID for the select menu."""

    placeholder: Optional[str] = None
    """Placeholder text if nothing is selected."""

    default_values: Optional[list[DefaultValue]] = field(default_factory=list)
    """
        List of default values for auto-populated select menu components. See [`DefaultValue`][scurrypy.parts.components.DefaultValue].
        Number of default values must be in the range of `min_values` to `max_values`.
    """

    min_values: Optional[int] = 1
    """Minimum number of items that must be chosen. Defaults to 1."""

    max_values: Optional[int] = 1
    """Maximum number of items that can be chosen. Defaults to 1."""

    required: Optional[bool] = False
    """Whether the select is required to answer in a modal. Defaults to `False`."""

    disabled: Optional[bool] = False
    """Whether select menu is disabled in a message. Defaults to `False`."""


@dataclass
class UserSelect(SelectMenu, ActionRowChild, LabelChild):
    """Represents the User Select component."""

    type: int = field(init=False, default=ComponentTypes.USER_SELECT)
    """Component type. Always `ComponentTypes.USER_SELECT` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

@dataclass
class RoleSelect(SelectMenu, ActionRowChild, LabelChild):
    """Represents the Role Select component."""

    type: int = field(init=False, default=ComponentTypes.ROLE_SELECT)
    """Component type. Always `ComponentTypes.ROLE_SELECT` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

@dataclass
class MentionableSelect(SelectMenu, ActionRowChild, LabelChild):
    """Represents the Mentionable Select component."""

    type: int = field(init=False, default=ComponentTypes.MENTIONABLE_SELECT)
    """Component type. Always `ComponentTypes.MENTIONABLE_SELECT` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""

@dataclass
class ChannelSelect(SelectMenu, ActionRowChild, LabelChild):
    """Represents the Channel Select component."""

    type: int = field(init=False, default=ComponentTypes.CHANNEL_SELECT)
    """Component type. Always `ComponentTypes.CHANNEL_SELECT` for this class. See [`ComponentTypes`][scurrypy.parts.components.ComponentTypes]."""



# scurrypy/core

from .error import DiscordError
from .model import DataModel
from .intents import Intents
from .permissions import Permissions
from .addon import Addon

__all__ = [
    "Addon",
    "DataModel",
    "DiscordError",
    "Intents",
    "Permissions"
]



class ActionRowChild: 
    """Marker class for all components that go into an action row.

    !!! tip "Children"
        [`Button`][scurrypy.parts.components.Button], 
        [`StringSelect`][scurrypy.parts.components.StringSelect], 
        [`UserSelect`][scurrypy.parts.components.UserSelect], 
        [`RoleSelect`][scurrypy.parts.components.RoleSelect], 
        [`MentionableSelect`][scurrypy.parts.components.MentionableSelect], 
        [`ChannelSelect`][scurrypy.parts.components.ChannelSelect]
    """
    ...

class SectionChild: 
    """Marker class for all components that go into a section.

    !!! tip "Children"
        [`TextDisplay`][scurrypy.parts.components_v2.TextDisplay]
    """
    ...

class SectionAccessoryChild: 
    """Marker class for all components that go into a section accessory.
    
    !!! tip "Children"
        [`Button`][scurrypy.parts.components.Button], 
        [`Thumbnail`][scurrypy.parts.components_v2.Thumbnail]
    """
    ...

class ContainerChild: 
    """Marker class for all components that go into a container.
    
    !!! tip "Children"
        [`ActionRowPart`][scurrypy.parts.components.ActionRowPart], 
        [`TextDisplay`][scurrypy.parts.components_v2.TextDisplay], 
        [`SectionPart`][scurrypy.parts.components_v2.SectionPart], 
        [`MediaGallery`][scurrypy.parts.components_v2.MediaGallery], 
        [`Separator`][scurrypy.parts.components_v2.Separator], 
        [`File`][scurrypy.parts.components_v2.File]
        [`FileUpload`][scurrypy.parts.components_v2.FileUpload]
    """
    ...

class LabelChild: 
    """Marker class for all components that go into a label.
    
    !!! tip "Children"
        [`TextInput`][scurrypy.parts.components.TextInput], 
        [`StringSelect`][scurrypy.parts.components.StringSelect], 
        [`UserSelect`][scurrypy.parts.components.UserSelect], 
        [`RoleSelect`][scurrypy.parts.components.RoleSelect], 
        [`MentionableSelect`][scurrypy.parts.components.MentionableSelect], 
        [`ChannelSelect`][scurrypy.parts.components.ChannelSelect], 
        [`FileUpload`][scurrypy.parts.components_v2.FileUpload]
    """
    ...

class ComponentTypes:
    ACTION_ROW = 1
    BUTTON = 2
    STRING_SELECT = 3
    TEXT_INPUT = 4
    USER_SELECT = 5
    ROLE_SELECT = 6
    MENTIONABLE_SELECT = 7
    CHANNEL_SELECT = 8
    SECTION = 9
    TEXT_DISPLAY = 10
    THUMBNAIL = 11
    MEDIA_GALLERY = 12
    FILE = 13
    SEPARATOR = 14
    CONTAINER = 17
    LABEL = 18
    FILE_UPLOAD = 19



from dataclasses import dataclass

from .base_resource import BaseResource

from ..models.emoji import EmojiModel
from ..models.message import MessageModel

from ..parts.message import MessagePart

@dataclass
class Message(BaseResource):
    """A Discord message."""

    id: int
    """ID of the message"""

    channel_id: int
    """Channel ID of the message."""

    async def fetch(self) -> MessageModel:
        """Fetches the message data based on the given channel ID and message id.

        Returns:
            (MessageModel): queried message
        """
        data = await self._http.request('GET', f"/channels/{self.channel_id}/messages/{self.id}")

        return MessageModel.from_dict(data)

    async def send(self, message: str | MessagePart) -> MessageModel:
        """Sends a new message to the current channel.

        Permissions:
            * `SEND_MESSAGES` → required to send your own messages

        Args:
            message (str | MessagePart): content as a string or MessagePart

        Returns:
            (MessageModel): created message
        """
        if isinstance(message, str):
            message = MessagePart(content=message)
        elif not message:
            raise ValueError("Missing message.")

        data = await self._http.request(
            "POST",
            f"/channels/{self.channel_id}/messages",
            data=message._prepare().to_dict(),
            files=[fp.path for fp in message.attachments] if message.attachments else None
        )
        return MessageModel.from_dict(data)

    async def edit(self, message: str | MessagePart) -> MessageModel:
        """Edits this message.

        Permissions:
            * `MANAGE_MESSAGES` → ONLY if editing another user's message

        Args:
            message (str | MessagePart): can be just text or the MessagePart for dynamic messages

        Returns:
            (MessageModel): updated message
        """
        if isinstance(message, str):
            message = MessagePart(content=message)
        elif not message:
            raise ValueError("Missing message.")

        data = await self._http.request(
            "PATCH", 
            f"/channels/{self.channel_id}/messages/{self.id}", 
            data=message._prepare().to_dict(),
            files=[fp.path for fp in message.attachments] if message.attachments else None)

        return MessageModel.from_dict(data)

    async def crosspost(self) -> MessageModel:
        """Crosspost this message in an Annoucement channel to all following channels.

        Permissions:
            * `SEND_MESSAGES` → required to publish your own messages
            * `MANAGE_MESSAGES` → required to publish messages from others

        Returns:
            (MessageModel): published (crossposted) message
        """
        data = await self._http.request('POST', f'/channels/{self.channel_id}/messages/{self.id}/crosspost')

        return MessageModel.from_dict(data)

    async def delete(self):
        """Deletes this message."""
        await self._http.request("DELETE", f"/channels/{self.channel_id}/messages/{self.id}")

    async def add_reaction(self, emoji: EmojiModel | str) -> None:
        """Add a reaction to this message.

        Permissions:
            * `READ_MESSAGE_HISTORY` → required to view message
            * `ADD_REACTIONS` → required to create reaction

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)
        elif not emoji:
            raise ValueError("Missing emoji.")

        await self._http.request(
            "PUT",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}/@me")
    
    async def remove_reaction(self, emoji: EmojiModel | str) -> None:
        """Remove the bot's reaction from this message.

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)
        elif not emoji:
            raise ValueError("Missing emoji.")

        await self._http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}/@me")

    async def remove_user_reaction(self, emoji: EmojiModel | str, user_id: int) -> None:
        """Remove a specific user's reaction from this message.

        Permissions:
            * `MANAGE_MESSAGES` → required to remove another user's reaction

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
            user_id (int): user's ID
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)
        elif not emoji:
            raise ValueError("Missing emoji.")

        await self._http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}/{user_id}")

    async def remove_all_reactions(self) -> None:
        """Clear all reactions from this message.

        Permissions:
            * `MANAGE_MESSAGES` → required to remove all reaction
        """
        await self._http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions")

    async def pin(self) -> None:
        """Pin this message to its channel's pins."""
        await self._http.request('PUT', f'/channels/{self.channel_id}/messages/pins/{self.id}')

    async def unpin(self) -> None:
        """Unpin this message from its channel's pins."""
        await self._http.request('DELETE', f'/channels/{self.channel_id}/messages/pins/{self.id}')



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

@dataclass
class AttachmentModel(DataModel):
    """Represents an attachment object."""

    id: int
    """Attachment ID."""

    filename: str
    """Name of the file."""

    title: Optional[str]
    """Title of the file."""

    description: Optional[str]
    """Description of the file."""

    content_type: Optional[str]
    """Media type of the file."""

    size: int
    """Size of file (in bytes)."""

    url: str
    """Source URL of the file."""

    proxy_url: str
    """A proxied URL of the file."""

    height: Optional[int]
    """Height of file (if image)."""

    width: Optional[int]
    """Width of file (if image)."""

    ephemeral: Optional[bool]
    """Whether this file is ephemeral."""

    duration_secs: Optional[float]
    """Duration of the file (for voice messages)."""

    waveform: Optional[str]
    """base64 encoded bytearray representing a sampled waveform (for voice messages)."""

    flags: Optional[int]
    """Attachment flags as a combined bitfield."""



# scurrypy/models

from .application import ApplicationFlags, ApplicationModel
from .attachment import AttachmentModel
from .channel import (
    ThreadMetadataModel,
    ChannelModel, 
    FollowedChannelModel,
    TagModel,
    DefaultReactionModel,
    ArchivedThreads
)
from .command import (
    ApplicationCommandTypes,
    ApplicationCommandOptionTypes,
    ApplicationCommandOptionChoiceModel,
    ApplicationCommandOptionModel,
    ApplicationCommandModel
)
from .emoji import EmojiModel
from .guild import ReadyGuildModel, GuildModel
from .integration import IntegrationModel
from .interaction import (
    InteractionCallbackDataModel, 
    InteractionCallbackModel,
    InteractionCallbackTypes,
    InteractionDataTypes,
    InteractionTypes,
    InteractionModel
)
from .message import MessageModel, PinnedMessageModel
from .role import RoleColorModel, RoleModel
from .user import UserModel, GuildMemberModel

__all__ = [
    "ApplicationFlags", "ApplicationModel",
    "AttachmentModel",
    "ChannelModel", "FollowedChannelModel", "ThreadMetadataModel", "TagModel", "DefaultReactionModel", "ArchivedThreads",
    "ApplicationCommandTypes", "ApplicationCommandOptionTypes", "ApplicationCommandOptionChoiceModel", 
    "ApplicationCommandOptionModel", "ApplicationCommandModel",
    "EmojiModel",
    "ReadyGuildModel", "GuildModel",
    "IntegrationModel",
    
    "InteractionCallbackDataModel", "InteractionCallbackModel", "InteractionCallbackTypes", 
    "InteractionDataTypes", "InteractionTypes", "InteractionModel",

    "MessageModel", "PinnedMessageModel",
    "RoleColorModel", "RoleModel",
    "UserModel", "GuildMemberModel"
]



"""
R = request
EP = endpoint
L = Lock
Q = Queue
H = header
B = Bucket
A + B = {A:B}

[R + EP]--|L|-->[Q + EP]--|send R|-->[add/update H + B]
    1. request by endpoint
    2. push request to queue with lock
    3. add/update header by bucket ID with send request

    * Queue by ENDPOINT/REQUEST
    * Bucket by HEADER
"""

import asyncio
import aiohttp
import aiofiles
import json
from typing import Any

from dataclasses import dataclass

from .error import DiscordError

import logging

logger = logging.getLogger(__name__)

@dataclass
class RequestItem:
    method: str
    endpoint: str
    data: dict = None
    params: dict = None
    files: dict = None
    future: asyncio.Future = None

@dataclass
class Bucket:
    remaining: int
    reset_after: float
    reset_on: float
    sleep_task: asyncio.Task = None

class HTTPClient:
    BASE = "https://discord.com/api/v10"
    MAX_RETRIES = 3

    def __init__(self):
        self.session = None

        # PRE-REQUEST
        self.queues: dict[str, asyncio.Queue] = {}  # maps EP -> Q
        self.queues_lock = asyncio.Lock() # locks queues dict for editing

        self.workers: dict[str, asyncio.Task] = {}  # maps EP -> worker

        # POST-REQUEST
        self.buckets: dict[str, Bucket] = {}  # maps B -> Bucket
        self.bucket_lock: dict[str, asyncio.Lock] = {} # maps B to Lock
        self.buckets_lock = asyncio.Lock() # locks buckets dict for editing

        self.global_lock = asyncio.Lock()
        self.global_reset = 0.0

    async def start(self, token: str):
        """Start the HTTP session."""

        if not self.session:
            self.session = aiohttp.ClientSession(headers={"Authorization": f"Bot {token}"})
            logger.info("HTTP session started.")
        else:
            logger.warning("HTTP session already initialized.")

    async def close(self):
        """Gracefully stop all workers and close the HTTP session."""

        if self.session: # just the session that needs to close!
            await self.session.close()
            logger.info("Session closed.")

    async def request(
        self,
        method: str,
        endpoint: str,
        *,
        data: dict | None = None,
        params: dict | None = None,
        files: Any | None = None,
    ):
        """Queue a request for the given endpoint.

        Args:
            method (str): HTTP method (e.g., POST, GET, DELETE, PATCH, etc.)
            endpoint (str): Discord endpoint (e.g., /channels/123/messages)
            data (dict | None, optional): relevant data
            params (dict | None, optional): relevant query params
            files (Any | None, optional): relevant files

        Returns:
            (Future | None): result or promise of request or None if failed
        """
        # ensure a queue is in place for the requested endpoint
        async with self.queues_lock:
            queue = self.queues.setdefault(endpoint, asyncio.Queue())

        if endpoint not in self.workers:
            self.workers[endpoint] = asyncio.create_task(self._worker(endpoint))

        # set promise
        future = asyncio.get_event_loop().create_future()

        def sanitize_query_params(params: dict | None) -> dict | None:
            """Sanitize a request's params for session.request

            Args:
                params (dict | None): query params (if any)

            Returns:
                (dict | None): the session.request-friendly version of params
            """
            if not params:
                return None
            return {k: ('true' if v is True else 'false' if v is False else v)
                for k, v in params.items() if v is not None}

        await queue.put(RequestItem(method, endpoint, data, sanitize_query_params(params), files, future))

        # return promise
        try:
            return await future
        except DiscordError as e:
            logger.error(e)
            return None

    async def _worker(self, endpoint: str):
        """Background worker that processes requests for this endpoint.

        Args:
            endpoint (str): the endpoint to receive requests
        """
        # fetch the queue by endpoint
        queue = self.queues[endpoint]

        while True:
            # get the next item in the queue
            item: RequestItem = await queue.get()

            if item is None: # sentinel = time to stop
                queue.task_done()
                break

            try:
                result = await self._send(item)
            except Exception as e:
                item.future.set_exception(e)
            else:
                item.future.set_result(result)
            finally:
                queue.task_done()

    async def _sleep_endpoint(self, endpoint: str, bucket: Bucket):
        """Let an endpoint sleep for the designated reset_after seconds.

        Args:
            endpoint (str): endpoint to sleep
            bucket (Bucket): endpoint's bucket info
        """
        logger.warning(f"Bucket {endpoint} rate limit is active. Sleeping for {bucket.reset_after}s...")
        await asyncio.sleep(bucket.reset_after)
        bucket.sleep_task = None
        logger.info(f"Bucket {endpoint} reset after {bucket.reset_after}s")

    async def _check_global_rate_limit(self):
        """Checks if the global rate limit is after now (active)."""
        now = asyncio.get_event_loop().time()
        if self.global_reset > now:
            async with self.global_lock:
                logger.warning(f"Global reset is active. Sleeping for {self.global_reset - now}s...")
                await asyncio.sleep(self.global_reset - now)
                logger.info(f"Global has reset after {self.global_reset - now}s...")

    async def _parse_response(self, resp: aiohttp.ClientResponse):
        """Parse the request's response for response details.

        Args:
            resp (aiohttp.ClientResponse): the response object

        Raises:
            DiscordError: Error object for pretty printing if an error is returned.

        Returns:
            (str | dict | None): request info (if any)
        """
        match resp.status:
            case 204:
                # No content
                return None

            case 200 | 201:
                # JSON body is guaranteed if successful
                try:
                    return await resp.json()
                except aiohttp.ContentTypeError:
                    return await resp.text()

            case _:
                # error handling
                try:
                    body = await resp.json()
                except aiohttp.ContentTypeError:
                    body = await resp.text()
                raise DiscordError(resp.status, body)
            
    async def _update_bucket_rate_limit(self, resp: aiohttp.ClientResponse, bucket_id: str, endpoint: str):
        """Update the bucket for this endpoint and sleep if necessary.

        Args:
            resp (aiohttp.ClientResponse): the response object
            bucket_id (str): bucket ID provided by Discord's headers
            endpoint (str): endpoint in which request was sent
        """
        # grab lock from dict of bucket locks with a lock on dict access
        async with self.buckets_lock:
            lock = self.bucket_lock.setdefault(bucket_id, asyncio.Lock())

        # update/add the bucket with Bucket lock
        async with lock:
            remaining = int(resp.headers.get('x-ratelimit-remaining', 1))
            reset_after = float(resp.headers.get('x-ratelimit-reset-after', 0))
            reset_on = float(resp.headers.get('x-ratelimit-reset', 0))

            bucket = self.buckets.get(bucket_id)

            if not bucket:
                bucket = Bucket(remaining, reset_after, reset_on)
                self.buckets[bucket_id] = bucket
            else:
                bucket.remaining = remaining
                bucket.reset_after = reset_after
                bucket.reset_on = reset_on

            logger.debug(f"[{endpoint}] {resp.method} bucket={bucket_id} reset_on={bucket.reset_on} remaining={bucket.remaining} reset_after={bucket.reset_after:.2f}s")

            if bucket.remaining == 0 and not bucket.sleep_task:
                bucket.sleep_task = asyncio.create_task(
                    self._sleep_endpoint(endpoint, bucket)
                )

            elif bucket.sleep_task and not bucket.sleep_task.done():
                await bucket.sleep_task

    async def _prepare_payload(self, item: RequestItem):
        """Prepares the payload based on `RequestItem`.

        Args:
            item (RequestItem): the request object

        Returns:
            (dict): kwargs to pass to session.request
        """
        if item.files and any(item.files):
            # payload = await self._make_payload(item.data, item.files)
            form = aiohttp.FormData()
            form.add_field("payload_json", json.dumps(item.data))

            for idx, file_path in enumerate(item.files):
                async with aiofiles.open(file_path, 'rb') as f:
                    f_data = await f.read()
                    form.add_field(
                        f'files[{idx}]',
                        f_data,
                        filename=file_path.split('/')[-1],
                        content_type='application/octet-stream'
                    )

            return {"data": form}

        return {"json": item.data}

    async def _send(self, item: RequestItem):
        """Core HTTP request executor.

        Args:
            item (RequestItem): request object

        Returns:
            (dict | str | None): Parsed JSON response if available, raw text if the
                response is not JSON, or None for HTTP 204 responses.
        """
        await self._check_global_rate_limit()

        kwargs = await self._prepare_payload(item)

        url = f"{self.BASE.rstrip('/')}/{item.endpoint.lstrip('/')}"
        
        async with self.session.request(
            method=item.method, url=url, params=item.params, timeout=15, **kwargs
        ) as resp:
            
            if resp.headers.get("X-RateLimit-Global") == "true":
                retry_after = float(resp.headers.get("Retry-After", 0))
                self.global_reset = asyncio.get_event_loop().time() + retry_after

            bucket_id = resp.headers.get('x-ratelimit-bucket')

            if bucket_id:
                await self._update_bucket_rate_limit(resp, bucket_id, item.endpoint)

            return await self._parse_response(resp)



from dataclasses import dataclass, field
from ..core.model import DataModel
from .base_event import Event

from typing import Optional

from ..models.interaction import InteractionModel

from ..models.user import UserModel
from ..models.user import GuildMemberModel
from ..models.role import RoleModel
from ..models.channel import ChannelModel
from ..models.message import MessageModel
from ..models.attachment import AttachmentModel

@dataclass
class ResolvedData(DataModel):
    """Represents the resolved data object."""

    users: Optional[dict[int, UserModel]]
    """Map of user snowflakes to user objects."""

    members: Optional[dict[int, GuildMemberModel]]
    """Map of member snowflakes to partial guild member objects.

    !!! note "Missing Fields"
        `user`, `deaf`, and `mute`.
    """

    roles: Optional[dict[int, RoleModel]]
    """Map of role snowflakes to role objects."""

    channels: Optional[dict[int, ChannelModel]]
    """Map of channel snowflakes to partial channel objects."""

    messages: Optional[dict[int, MessageModel]]
    """Map of message snowflakes to partial message objects."""

    attachments: Optional[dict[int, AttachmentModel]]
    """Map of attachment snowflakes to attachment objects."""

# ----- Command Interaction -----

@dataclass
class ApplicationCommandOptionData(DataModel):
    """Represents the response options from a slash command."""
    
    name: str
    """Name of the command option."""

    type: int
    """Type of command option. See [`CommandOptionTypes`][scurrypy.parts.command.CommandOptionTypes]."""

    value: str
    """
    Raw value from Discord as a string.
    
    Convert based on option type:
        - INTEGER/USER/CHANNEL/ROLE/ATTACHMENT: int(value)
        - NUMBER: float(value)  
        - BOOLEAN: value.lower() == 'true'
        - STRING: value (no conversion needed)
    """

    focused: bool
    """Whether this option is the currently focused option for autocomplete."""

@dataclass
class ApplicationCommandData(DataModel):
    """Represents the response from a command."""

    id: int
    """ID of the command."""

    name: str
    """Name of the command."""
    
    type: int
    """Type of command (e.g., message, user, slash)."""

    guild_id: Optional[int]
    """ID of guild from which the command was invoked."""

    target_id: Optional[int]
    """ID of the user or message from which the command was invoked (message/user commands only)."""

    resolved: Optional[ResolvedData]
    """Converted users + roles + channels + attachments."""

    options: Optional[list[ApplicationCommandOptionData]] = field(default_factory=list)
    """Options of the command (slash command only)."""

    def get_focused_value(self):
        opt = next((opt for opt in self.options if opt), None)
        
        return opt.value if opt else ""

    def get_option(self, option_name: str, default = None):
        """Get the input for a command option by name.

        Args:
            option_name (str): option to fetch input from

        Raises:
            ValueError: invalid option name

        Returns:
            (str | int | float | bool): input data of specified option
        """
        for option in self.options:
            if option.name == option_name:
                return option.value
        
        if default is not None:
            return default
        
        raise ValueError(f"Option '{option_name}' not found")

# ----- Component Interaction -----

@dataclass
class MessageComponentData(DataModel):
    """Represents the select response from a select component."""

    custom_id: str
    """Unique ID associated with the component."""

    component_type: int
    """Type of component."""

    resolved: Optional[ResolvedData]
    """Resolved entities from selected options."""

    values: Optional[list[str]] = field(default_factory=list)
    """Select values (if any)."""

# ----- Modal Interaction -----

@dataclass
class ModalComponentData(DataModel):
    """Represents the modal field response from a modal."""

    type: int
    """Type of component."""
    
    value: Optional[str]
    """Text input value (Text Input component only)."""

    custom_id: str
    """Unique ID associated with the component."""

    values: Optional[list[str]] = field(default_factory=list)
    """String select values (String Select component only)."""

@dataclass
class ModalComponent(DataModel):
    """Represents the modal component response from a modal."""

    type: int
    """Type of component."""

    component: ModalComponentData
    """Data associated with the component."""

@dataclass
class ModalData(DataModel):
    """Represents the modal response from a modal."""
    
    custom_id: str
    """Unique ID associated with the modal."""

    data: ResolvedData
    """Resolved entities from modal data."""

    components: list[ModalComponent] = field(default_factory=list)
    """Components on the modal."""

    def get_modal_data(self, custom_id: str):
        """Fetch a modal field's data by its custom ID

        Args:
            custom_id (str): custom ID of field to fetch

        Raises:
            ValueError: invalid custom ID

        Returns:
            (str | list[str]): component values (if string select) or value (if text input)
        """
        from ..parts.components import ComponentTypes
        
        for component in self.components:
            if custom_id != component.component.custom_id:
                continue

            t = component.component.type

            if t in [
                ComponentTypes.STRING_SELECT, 
                ComponentTypes.USER_SELECT, 
                ComponentTypes.ROLE_SELECT, 
                ComponentTypes.MENTIONABLE_SELECT, 
                ComponentTypes.CHANNEL_SELECT,
                ComponentTypes.FILE_UPLOAD      # select menus (w. possibly many option selects!)
            ]:
                return component.component.values
            
            # text input
            return component.component.value

        raise ValueError(f"Component custom ID '{custom_id}' not found.")

@dataclass
class InteractionEvent(Event, InteractionModel):
    """Represents the interaction response."""

    data: Optional[ApplicationCommandData | MessageComponentData | ModalData]
    """Interaction response data."""

    @classmethod
    def from_dict(cls, data: dict):
        from ..models.interaction import InteractionTypes
        
        obj = super().from_dict(data) # InteractionModel's DataModel

        interaction_data = data.get("data")
        interaction_type = data.get("type")

        match interaction_type:
            case InteractionTypes.APPLICATION_COMMAND | InteractionTypes.APPLICATION_COMMAND_AUTOCOMPLETE:
                obj.data = ApplicationCommandData.from_dict(interaction_data)
            case InteractionTypes.MESSAGE_COMPONENT:
                obj.data = MessageComponentData.from_dict(interaction_data)
            case InteractionTypes.MODAL_SUBMIT:
                obj.data = ModalData.from_dict(interaction_data)

        return obj



from dataclasses import dataclass, field
from ..core.model import DataModel

from .components_v2 import Label

@dataclass
class ModalPart(DataModel):
    """Represents the Modal object."""

    title: str = None
    """Title of the popup modal."""

    custom_id: str = None
    """ID for the modal."""

    components: list[Label] = field(default_factory=list)
    """1 to 5 components that make up the modal."""



from typing import TypedDict, Unpack

class IntentFlagParams(TypedDict, total=False):
    """Gateway intent selection parameters.
    !!! important
        Param name MUST match the intent it represents! 
        For example, "GUILD_MESSAGES" is "guild_messages".
    """
    guilds: bool
    guild_members: bool
    guild_emojis_and_stickers: bool
    guild_integrations: bool
    guild_webhooks: bool
    guild_messages: bool
    guild_message_reactions: bool
    message_content: bool

class Intents:
    """Gateway intent flags (bitwise).  
        For an exhaustive list what intents let your bot listen to what events, 
        see <a href="https://discord.com/developers/docs/events/gateway#list-of-intents" target="_blank" rel="noopener">list of intents</a>.

    !!! note
        Not all intents are listed. Intents not listed are not yet supported.
    """

    GUILDS = 1 << 0
    """Receive events related to guilds."""

    GUILD_MEMBERS = 1 << 1
    """
    !!! warning "Privileged Intent"
        Requires the app setting `Server Members Intent` to be toggled.
    Receive events related to guild members.
    """

    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    """Receive events related to custom emojis and stickers."""

    GUILD_INTEGRATIONS = 1 << 4
    """Receive events related to integrations within a guild."""

    GUILD_WEBHOOKS = 1 << 5
    """Track webhook events within a guild."""

    GUILD_MESSAGES = 1 << 9
    """Receive events about messages within a guild."""

    GUILD_MESSAGE_REACTIONS = 1 << 10
    """Track changes in reactions on messages."""

    MESSAGE_CONTENT = 1 << 15
    """
    !!! warning "Privileged Intent"
        Requires the app setting `Message Content Intent` to be toggled.
    Access content of messages.
    """

    DEFAULT = GUILDS | GUILD_MESSAGES

    @staticmethod
    def set(**flags: Unpack[IntentFlagParams]):
        """Set bot intents. See [`Intents`][scurrypy.core.intents.Intents].
        `Intents.DEFAULT` will also be set.

        Args:
            **flags (Unpack[IntentFlagParams]): intents to set

        Raises:
            (ValueError): invalid flag

        Returns:
            (int): combined intents field
        """
        intents = Intents.DEFAULT
        for k, v in flags.items():
            if v:
                try:
                    intents |= getattr(Intents, k.upper())
                except AttributeError:
                    raise ValueError(f"Unknown intent flag: '{k}'")
        
        return intents

    @staticmethod
    def has(intents: int, intent: int):
        """Checks if an intent bit is toggled.

        Args:
            intents (int): bot intents
            intent (int): intent bit to check

        Returns:
            (bool): if intent bit is toggled
        """
        return (intents & intent) == intent



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

from typing import Optional

from ..models.user import GuildMemberModel
from ..models.emoji import EmojiModel

class ReactionType:
    """Reaction types."""

    NORMAL = 0
    """A standard emoji."""

    BURST = 1
    """A super emoji."""

@dataclass
class ReactionAddEvent(Event, DataModel):
    """Reaction added event."""

    type: int
    """Type of reaction added."""

    user_id: int
    """ID of user who added the emoji."""

    emoji: EmojiModel
    """Emoji used to react."""

    channel_id: int
    """ID of the channel where the reaction took place."""

    message_id: int
    """ID of the message where the reaction took place."""

    guild_id: Optional[int]
    """ID of the guild where the reaction took place (if in a guild)."""

    burst: bool
    """Whether the emoji is super."""

    member: Optional[GuildMemberModel]
    """Partial member object of the guild member that added the emoji (if in a guild)."""

    message_author_id: Optional[int]
    """ID of the user who sent the message where the reaction was added."""

@dataclass
class ReactionRemoveEvent(Event, DataModel):
    """Reaction removed event."""

    type: int
    """Type of reaction removed."""

    user_id: int
    """ID of user who removed their reaction."""

    emoji: EmojiModel
    """Emoji data of the emoji where the reaction was removed."""

    channel_id: int
    """ID of the channel where the reaction was removed."""

    message_id: int
    """ID of the message where the reaction was removed."""

    guild_id: Optional[int]
    """ID of the guild where the reaction was removed (if in a guild)."""

    burst: bool
    """If the emoji of the removed reaction is super."""

class ReactionRemoveAllEvent(Event, DataModel):
    """Remove all reactions event."""

    channel_id: int
    """ID of the channel where all reaction were removed."""

    message_id: int
    """ID of the message where all reaction were removed."""

    guild_id: Optional[int]
    """ID of the guild where all reaction were removed (if in a guild)."""

@dataclass
class ReactionRemoveEmojiEvent(Event, DataModel):
    """All reactions of a specific emoji removed."""

    emoji: EmojiModel
    """Emoji data of the removed reaction emoji."""

    channel_id: int
    """ID of the channel where the reaction emoji was removed."""

    message_id: int
    """ID of the message where the reaction emoji was removed."""

    guild_id: Optional[int]
    """ID of the guild where the reaction emoji was removed. (if in a guild)"""



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

from .application import ApplicationModel

@dataclass
class IntegrationModel(DataModel):
    """Represents a guild integration."""

    id: int
    """ID of the integration."""

    name: str
    """Name of the integration."""

    type: str
    """Type of integration (e.g.,`'twitch'`, `'youtube'`, `'discord'`, or `'guild_subscription'`)."""

    enabled: bool
    """If the integration is enabled."""

    application: Optional[ApplicationModel]
    """The bot application for Discord integrations."""



from dataclasses import dataclass
from ..core.model import DataModel
from typing import Optional
from .base_event import Event

from ..models.user import GuildMemberModel
from ..models.channel import ChannelModel
from ..models.guild import UnavailableGuild, GuildModel

@dataclass
class GuildCreateEvent(Event, GuildModel):
    """Received when the bot has joined a guild."""
    
    joined_at: str
    """ISO8601 timestamp of when app joined the guild."""

    large: bool
    """If the guild is considered large."""

    member_count: int
    """Total number of members in the guild."""

    members: list[GuildMemberModel]
    """Users in the guild."""

    channels: list[ChannelModel]
    """Channels in the guild."""

    threads: list[ChannelModel]
    """All active threads in the guild that are viewable."""

    unavailable: Optional[bool]
    """`True` if the guild is unavailable due to an outage."""

@dataclass
class GuildUpdateEvent(Event, DataModel):
    """Received when a guild has been edited."""

    id: int
    """ID of the guild."""

    name: str
    """Name of the guild."""

    icon: str
    """Image hash of the guild's icon."""

    description: str
    """Description of the guild."""

    banner: str
    """Image hash of the guild's banner."""

    joined_at: str
    """ISO8601 timestamp of when app joined the guild."""

    large: bool
    """If the guild is considered large."""

    member_count: int
    """Total number of members in the guild."""

@dataclass
class GuildDeleteEvent(Event, UnavailableGuild):
    """Received when the bot has left a guild or the guild was deleted."""
    pass

@dataclass
class GuildMemberAddEvent(Event, GuildMemberModel):
    """Received when a member joins a guild the bot is in.

    !!! warning
        Requires privileged `GUILD_MEMBERS` intent.
    """

    guild_id: int
    """ID of the guild."""


from ..models.user import UserModel

@dataclass
class GuildMemberUpdateEvent(Event, DataModel):
    """Received when a guild member is updated.
    
    !!! warning
        Requires privileged `GUILD_MEMBERS` intent.
    """
    guild_id: int
    """ID of the guild."""

    roles: list[int]
    """List of user's roles (their IDs)."""

    user: UserModel
    """The User object."""

    avatar: str
    """Guild avatar hash."""

    banner: str
    """Guild banner hash."""

    joined_at: str
    """When the user joined the guild"""

@dataclass
class GuildMemberRemoveEvent(Event, DataModel):
    """Received when a member leaves or is kicked/banned from a guild the bot is in.
    
    !!! warning
        Requires privileged `GUILD_MEMBERS` intent.
    """

    guild_id: int
    """ID of the guild."""

    user: UserModel
    """User object of the user leaving the guild."""


from ..models.emoji import EmojiModel

@dataclass
class GuildEmojisUpdateEvent(Event, DataModel):
    """Received when a guild updates their emojis."""

    guild_id: int
    """ID of the guild."""

    emojis: list[EmojiModel]
    """Complete set of guild emojis with changes."""



from dataclasses import dataclass
from ..core.model import DataModel

@dataclass
class ImageData(DataModel):
    """Represents Discord's data URI scheme for images."""
    
    path: str = None
    """Path to image."""

    @property
    def uri(self):
        """Creates the Base64 data URI scheme.

        Raises:
            ValueError: Unknown file type

        Returns:
            (str): the formatted data URI scheme
        """
        import base64, mimetypes

        mime, _ = mimetypes.guess_type(self.path)
        if mime is None:
            raise ValueError("Unknown file type.")

        with open(self.path, "rb") as f:
            encoded = base64.b64encode(f.read()).decode()

        return f"data:{mime};base64,{encoded}"



from dataclasses import dataclass

from .base_resource import BaseResource

from ..models.emoji import EmojiModel

from ..parts.guild_emoji import *

@dataclass
class GuildEmoji(BaseResource):
    """Represents a Discord Guild Emoji."""

    guild_id: int
    """Guild ID of the emojis."""

    async def fetch(self, emoji_id: int) -> EmojiModel:
        """Fetch an emoji from this guild.

        Args:
            emoji_id (int): emoji ID

        Returns:
            (EmojiModel): queried guild emoji
        """
        data = await self._http.request("GET", f"/guilds/{self.guild_id}/emojis/{emoji_id}")

        return EmojiModel.from_dict(data)
    
    async def fetch_all(self) -> list[EmojiModel]:
        """Fetch all emojis from this guild.

        Returns:
            (list[EmojiModel]): queried list of guild emojis
        """
        data = await self._http.request("GET", f"/guilds/{self.guild_id}/emojis")

        return [EmojiModel.from_dict(emoji) for emoji in data]

    async def create(self, params: CreateGuildEmoji) -> EmojiModel:
        """Create a new emoji for this guild.

        Args:
            params (CreateGuildEmoji): fields for creating a guild emoji

        Returns:
            (EmojiModel): new emoji
        """

        data = await self._http.request(
            'POST', 
            f'/guilds/{self.guild_id}/emojis', 
            data=params.to_dict()
        )

        return EmojiModel.from_dict(data)
    
    async def edit(self, emoji_id: int, params: EditGuildEmoji) -> EmojiModel:
        """Edit a guild emoji in this guild.

        Args:
            emoji_id (int): ID of the emoji to edit
            params (EditGuildEmoji): params for editing a guild's emoji

        Returns:
            (EmojiModel): updated emoji
        """

        data = await self._http.request(
            'PATCH', 
            f'/guilds/{self.guild_id}/emojis/{emoji_id}', 
            data=params.to_dict()
        )

        return EmojiModel.from_dict(data)

    async def delete(self, emoji_id: int) -> None:
        """Delete an emoji from this guild.

        Permissions:
            * `CREATE_GUILD_EXPRESSIONS` → if created by the current user (or `MANAGE_GUILD_EXPRESSIONS`)
            * `MANAGE_GUILD_EXPRESSIONS` → for other emojis

        Args:
            emoji_id (int): ID of the emoji
        """

        await self._http.request('DELETE', f'/guilds/{self.guild_id}/emojis/{emoji_id}')



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

from ..models.role import RoleModel

@dataclass
class RoleCreateEvent(Event, DataModel):
    """Received when a guild role is created."""

    guild_id: int
    """Guild ID of the role."""

    role: RoleModel
    """The new role."""

@dataclass
class RoleUpdateEvent(Event, DataModel):
    """Received when a guild role is updated."""

    guild_id: int
    """Guild ID of the role."""

    role: RoleModel
    """The new role."""

@dataclass
class RoleDeleteEvent(Event, DataModel):
    """Received when a guild role is deleted."""

    guild_id: int
    """Guild ID of the role."""

    role_id: int
    """Role ID of the role."""



import asyncio
import json
import websockets

import logging

logger = logging.getLogger(__name__)

MIN_BACKOFF = 5

class GatewayClient:
    def __init__(self, gateway_url: str, shard_id: int, total_shards: int):
        """Initialize this websocket.

        Args:
            gateway_url (str): gateway URL provided by GET /gateway/bot endpoint
            shard_id (int): assigned shard ID
            total_shards (int): total shard count provided by GET /gateway/bot endpoint
        """
        self.shard_id = shard_id
        self.total_shards = total_shards
        self.ws = None
        self.seq = None
        self.session_id = None
        self.allow_resume = False # default: assume IDENTIFY
        self.backoff = MIN_BACKOFF
        self.heartbeat_task = None
        self.heartbeat_interval = None
        self.event_queue = asyncio.Queue()

        self.base_url = gateway_url
        self.url_params = "?v=10&encoding=json"

    async def wait_reconnect(self):
        """Sleep for exponentially increasing time between reconnects."""
        
        logger.warning(f"SHARD ID {self.shard_id}: Disconnected, reconnecting in {self.backoff}s...")
        await asyncio.sleep(self.backoff)

        self.backoff = min(self.backoff * 2, 60)

    async def start(self, token: str, intents: int):
        """Start this websocket's connection.

        Args:
            token (str): the bot's token
            intents (int): the bot's intents
        """
        while True:
            try:
                await self.connect_ws()

                if self.allow_resume and self.session_id and self.seq is not None:
                    logger.debug(f"SHARD ID {self.shard_id}: Attempting to resume...")
                    await self.resume(token)
                else:
                    logger.debug(f"SHARD ID {self.shard_id}: Attempting to identify...")
                    await self.identify(token, intents)

                await self._listen()  # blocks until disconnect

            except websockets.exceptions.ConnectionClosedOK:
                logger.info(f"SHARD ID {self.shard_id}: Connection closed properly.")

                break

            except (ConnectionError, websockets.exceptions.ConnectionClosedError) as e:
                logger.error(f"SHARD ID {self.shard_id}: {e}")

                await self.close_ws()
                await self.wait_reconnect()

            except Exception:
                logger.exception(f"SHARD ID {self.shard_id}: Unexpected error")
                
                await self.close_ws()
                await self.wait_reconnect()

    async def connect_ws(self):
        """Connect to Discord's Gateway (websocket)."""

        # connect to websocket
        self.ws = await websockets.connect(self.base_url + self.url_params)
        logger.info(f"SHARD ID {self.shard_id}: Connected to Discord!")

        # wait to recv HELLO
        hello = await self.receive()

        # extra info from recv'd HELLO
        self.heartbeat_interval = hello["d"]["heartbeat_interval"] / 1000

        # start heartbeat in background
        self.heartbeat_task = asyncio.create_task(self.heartbeat())

    async def send(self, data: dict):
        """Send data through the websocket.

        Args:
            data (dict): data to send
        """
        await self.ws.send(json.dumps(data))

    async def receive(self):
        """Receive data through the websocket.

        Returns:
            (dict): websocket data
        """
        return json.loads(await self.ws.recv())

    async def heartbeat(self):
        """Heartbeat task to keep connection alive."""

        # add jitter only on before the first heartbeat
        import random
        jitter = random.uniform(0, 1)
        await asyncio.sleep(self.heartbeat_interval * jitter)

        while self.ws:
            await self.send({"op": 1, "d": self.seq})
            logger.debug(f"SHARD ID {self.shard_id}: Heartbeat sent")
            await asyncio.sleep(self.heartbeat_interval)

    async def identify(self, token: str, intents: int):
        """Send an IDENTIFY payload to handshake for bot.

        Args:
            token (str): the bot's token
            intents (int): the bot's intents

        Raises:
            (ConnectionError): invalid intents
        """

        if not isinstance(intents, int):
            raise ConnectionError("Invalid intents.")

        await self.send({
            'op': 2,
            'd': {
                'token': f"Bot {token}",
                'intents': intents,
                'properties': {
                    'os': 'my_os',
                    'browser': 'my_browser',
                    'device': 'my_device'
                },
                'shards': [self.shard_id, self.total_shards]
            }
        })
        logger.info(f"SHARD ID {self.shard_id}: IDENIFY Sent.")

    async def resume(self, token: str):
        """Send a RESUME payload to resume a connection.

        Args:
            token (str): the bot's token
        """

        await self.send({
            'op': 6,
            'd': {
                'token': f"Bot {token}",
                'session_id': self.session_id,
                'seq': self.seq
            }
        })
        logger.info(f"SHARD ID {self.shard_id}: Resume Sent.")

    async def _listen(self):
        """Listen for events and queue them to be consumed by Client.

        Raises:
            (ConnectionError): an error occurred
        """

        while self.ws:
            data = await self.receive()
            op_code = data.get("op")

            match op_code:
                case 0:  # DISPATCH
                    if data.get('s') is not None:
                        self.seq = data['s']

                    event_data = data.get('d')
                    dispatcher_type = data.get("t")

                    if dispatcher_type == "READY":
                        self.session_id = event_data.get("session_id")
                        self.base_url = event_data.get("resume_gateway_url", self.base_url)
                        self.backoff = MIN_BACKOFF
                        
                    elif dispatcher_type == "RESUMED":
                        self.backoff = MIN_BACKOFF

                    await self.event_queue.put((dispatcher_type, event_data))

                case 7:  # RECONNECT
                    self.allow_resume = True
                    logger.debug(f"SHARD ID {self.shard_id}: Reconnect requested by server.")

                    raise ConnectionError("Reconnect requested by server.")

                case 9: # INVALID_SESSION
                    resumable = bool(data.get("d"))
                    self.allow_resume = resumable

                    if resumable:
                        logger.debug(f"SHARD ID {self.shard_id}: Invalid session (resumable).")
                    else:
                        self.session_id = self.seq = None
                        logger.debug(f"SHARD ID {self.shard_id}: Invalid session (not resumable).")

                    raise ConnectionError("Invalid session.")

                case 11:  # HEARTBEAT_ACK
                    logger.debug(f"SHARD ID {self.shard_id}: Heartbeat ACK")

    async def close_ws(self):
        """Close the websocket connection if one is still open and cancels heartbeat."""

        logger.info(f"Shard ID {self.shard_id}: Closing connection...")
        if self.ws:
            if self.heartbeat_task:
                self.heartbeat_task.cancel()
                try:
                    await self.heartbeat_task
                except asyncio.CancelledError:
                    pass
                self.heartbeat_task = None
            await self.ws.close()

        self.ws = None



from dataclasses import dataclass, field
from ..core.model import DataModel

from typing import Optional, TypedDict, Unpack

from .embed import EmbedPart
from .components import ActionRowPart
from .components_v2 import ContainerPart

class MessageFlags:
    """Flags that can be applied to a message."""

    CROSSPOSTED = 1 << 0
    """Message has been published."""

    IS_CROSSPOST = 1 << 1
    """Message originated from another channel."""

    SUPPRESS_EMBEDS = 1 << 2
    """Hide embeds (if any)."""

    EPHEMERAL = 1 << 6
    """Only visible to the invoking user."""

    LOADING = 1 << 7
    """Thinking response."""

    IS_COMPONENTS_V2 = 1 << 15
    """This message includes Discord's V2 Components."""

class MessageFlagParams(TypedDict, total=False):
    """Parameters for setting message flags. See [`MessageFlags`][scurrypy.parts.message.MessageFlags]."""
    crossposted: bool
    is_crosspost: bool
    suppress_embeds: bool
    ephemeral: bool
    loading: bool
    is_components_v2: bool

class MessageReferenceTypes:
    """Constants associated with how reference data is populated."""

    DEFAULT = 0
    """Standard reference used by replies."""

    FORWARD = 1
    """Reference used to point to a message at a point in time."""

@dataclass
class MessageReference(DataModel):
    """Represents the Message Reference object."""

    message_id: int = None
    """ID of the originating message."""

    channel_id: int = None
    """
        Channel ID of the originating message.
        !!! note
            Optional for default type, but REQUIRED for forwards.
    """

    type: int = MessageReferenceTypes.DEFAULT
    """Type of reference. Defaults to `DEFAULT`. See [`MessageReferenceTypes`][scurrypy.parts.message.MessageReferenceTypes]."""

@dataclass
class Attachment(DataModel):
    """Represents an attachment."""

    id: int = field(init=False, default=None)
    """ID of the attachment (internally set)."""

    path: str = None
    """Relative path to the file."""

    description: str = None
    """Description of the file."""

    def to_dict(self):
        return {
            'id': self.id,
            'filename': self.path.split('/')[-1],
            'description': self.description
        }

@dataclass
class MessagePart(DataModel):
    """Describes expected params when editing/creating a message."""

    content: Optional[str] = None
    """Message text content."""

    flags: Optional[int] = 0
    """Message flags. See [`MessageFlags`][scurrypy.parts.message.MessageFlags].

    !!! note
        Flags are ignored if editing an existing message.
    """

    components: Optional[list[ActionRowPart | ContainerPart]] = field(default_factory=list)
    """Components to be attached to this message."""

    attachments: Optional[list[Attachment]] = field(default_factory=list)
    """Attachments to be attached to this message."""

    embeds: Optional[list[EmbedPart]] = field(default_factory=list)
    """Embeds to be attached to this message."""

    message_reference: Optional[MessageReference] = None
    """Message reference if reply."""

    def _prepare(self):
        """Prepares MessagePart for ANY internally set attributes.

        Returns:
            (MessagePart): self
        """
        # set attachment IDs (if any)
        if self.attachments:
            for idx, file in enumerate(self.attachments):
                file.id = idx
        
        return self

    def set_flags(self, **flags: Unpack[MessageFlagParams]):
        """Set this message's flags using MessageFlagParams.

        Args:
            **flags (Unpack[MessageFlagParams]): message flags to set. (set respective flag to True to toggle.)

        Raises:
            (ValueError): invalid flag

        Returns:
            (MessagePart): self
        """
        _flag_map = {
            'crossposted': MessageFlags.CROSSPOSTED,
            'is_crosspost': MessageFlags.IS_CROSSPOST,
            'suppress_embeds': MessageFlags.SUPPRESS_EMBEDS,
            'ephemeral': MessageFlags.EPHEMERAL,
            'loading': MessageFlags.LOADING,
            'is_components_v2': MessageFlags.IS_COMPONENTS_V2,
        }

        # each flag maps to a specific bit position!
        for name, value in flags.items():
            if name not in _flag_map:
                raise ValueError(f"Invalid flag: {name}")
            if value:
                self.flags |= _flag_map[name]
                
        return self



from dataclasses import dataclass
from ..core.model import DataModel
from ..core.permissions import Permissions

from typing import Optional

@dataclass
class ThreadMetadataModel(DataModel):
    """Represents the thread metadata object."""

    archived: bool
    """Whether the thread is archived."""

    auto_archive_duration: int
    """How long to wait until the thread is hidden (in minutes)."""

    archive_timestamp: str
    """ISO8601 timestamp of when the thread's archive status was last changed."""

    locked: bool
    """Whether the thread is locked.
    
    !!! note
        Only users with `MANAGE_THREADS` can unarchive the thread.
    """

    invitable: Optional[bool]
    """Whether non-moderators can add other non-moderators to the thread (private threads only)."""

    create_timestamp: Optional[str]
    """ISO8601 timestamp of thread creation (field only exists after Jan 09, 2022)."""

from .user import GuildMemberModel

@dataclass
class ThreadMemberModel(DataModel):
    """Represents a user that has joined a thread."""

    id: Optional[int]
    """ID of the thread."""

    user_id: Optional[int]
    """ID of the user."""

    join_timestamp: str
    """ISO8601 timestamp of when the user last joined the thread."""

    member: Optional[GuildMemberModel]
    """Additional information about the user.
    
    !!! note
        Only present when `with_member` is toggled on request.
    """

@dataclass
class TagModel(DataModel):
    """Represents the tag object found in `GUILD_FORUM` channels."""
    
    id: int
    """ID of the tag."""

    name: str
    """Name of the tag."""

    moderated: bool
    """Whether the tag can only be added/removed by a member with `MANAGE_THREADS`."""
    
    emoji_id: int
    """ID of a guild's custom emoji."""
    
    emoji_name: str
    """Unicode character of the emoji."""

@dataclass
class DefaultReactionModel(DataModel):
    """Represents the default reaction for a `GUILD_FORUM` post."""

    emoji_id: int
    """ID of the guild's custom emoji."""

    emoji_name: str
    """Unicode character of the emoji."""

from .user import UserModel

@dataclass
class ChannelModel(DataModel):
    """Represents a Discord guild channel."""

    id: int
    """ID of the channel."""

    type: int
    """Type of channel."""

    guild_id: Optional[int]
    """Guild ID of the channel."""

    parent_id: Optional[int]
    """Category ID of the channel."""

    position: Optional[int]
    """Position of the channel."""

    name: Optional[str]
    """Name of the channel."""

    topic: Optional[str]
    """Topic of the channel."""

    nsfw: Optional[bool]
    """If the channel is flagged NSFW."""

    last_message_id: Optional[int]
    """ID of the last message sent in the channel."""

    rate_limit_per_user: Optional[int]
    """Seconds user must wait between sending messages in the channel."""

    recipients: Optional[list[UserModel]]
    """Recipients of the DM."""

    icon: Optional[str]
    """Icon hash of the group DM."""

    owner_id: Optional[int]
    """ID of the creator of the group DM or thread."""

    application_id: Optional[int]
    """ID of the application that created the DM or thread."""

    last_pin_timestamp: Optional[str]
    """ISO8601 timestamp of the last pinned messsage in the channel."""

    permissions: Optional[int]
    """Permissions for the invoking user in this channel.
        Includes role and overwrite calculations. [`INT_LIMIT`]
    """

    thread_metadata: Optional[ThreadMetadataModel]
    """Thread-specific fields not needed by other channels."""

    member: Optional[ThreadMemberModel]
    """Thread member object for the current user if they have joined the thread."""

    default_auto_archive_duration: Optional[int]
    """Default duration in minutes threads will be hidden after period of inactivity."""

    flags: Optional[int]
    """Channel flags combined as a bitfield. See [ChannelFlags][scurrypy.parts.channel.ChannelFlags]."""

    available_tags: Optional[list[TagModel]]
    """Set of tags that can be applied to a `GUILD_FORUM` post."""

    applied_tags: Optional[list[int]]
    """Set of tags applied to a `GUILD_FORUM` post."""

    default_reaction_emoji: Optional[DefaultReactionModel]
    """Emoji to show in the add reaction button in a `GUILD_FORUM` post."""

    default_thread_rate_limit_per_user: Optional[int]
    """Rate limit per user set on newly created threads.
    
    !!! note
        This field does not live update!
    """

    default_sort_order: Optional[int]
    """Default forum sort order. See [`SortOrderTypes`][scurrypy.parts.channel.SortOrderTypes]. Defaults to `None`."""

    default_forum_layout: Optional[int]
    """Default forum layout view. See [`ForumLayoutTypes`][scurrypy.parts.channel.ForumLayoutTypes]. Defaults to `ForumLayoutTypes.NOT_SET`."""

    def user_can(self, permission_bit: int):
        """Checks `permissions` to see if permission bit is toggled.

        !!! warning
            If `permission` field is `None`, this function always returns `False`.

        Args:
            permission_bit (int): permission bit. See [Permissions][scurrypy.core.permissions.Permissions].

        Returns:
            (bool): whether the user has this permission
        """
        if not self.permissions:
            return False
        return Permissions.has(self.permissions, permission_bit)

@dataclass
class FollowedChannelModel(DataModel):
    """Represents the followed channel object."""

    channel_id: int
    """ID of the source channel."""

    webhook_id: int
    """Target webhook ID created."""

@dataclass
class ArchivedThreads(DataModel):
    """Response body for fetching archived threads."""

    threads: list[ChannelModel]
    """The archived threads."""

    members: list[ThreadMemberModel]
    """Thread member for each returned thread the bot has joined."""

    has_more: bool
    """Whether there are additional threads to be returned with subsequent calls."""



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

@dataclass
class UserModel(DataModel):
    """Describes the User object."""

    id: int
    """ID of the user."""

    username: str
    """Username of the user."""

    discriminator: str
    """Discriminator of the user (#XXXX)"""

    global_name: str
    """Global name of the user."""

    avatar: str
    """Image hash of the user's avatar."""

    bot: Optional[bool]
    """If the user is a bot."""

    system: Optional[bool]
    """If the user belongs to an OAuth2 application."""

    mfa_enabled: Optional[bool]
    """Whether the user has two factor enabled."""

    banner: Optional[str]
    """Image hash of the user's banner."""

    accent_color: Optional[int]
    """Color of user's banner represented as an integer."""

    locale: Optional[str]
    """Chosen language option of the user."""

@dataclass
class GuildMemberModel(DataModel):
    """Represents a guild member."""

    roles: list[int]
    """List of roles registered to the guild member."""

    user: UserModel
    """User data associated with the guild member."""

    nick: str
    """Server nickname of the guild member."""

    avatar: str
    """Server avatar hash of the guild mmeber."""

    joined_at: str
    """ISO8601 timestamp of when the guild member joined server."""

    deaf: bool
    """If the member is deaf in a VC (input)."""

    mute: bool
    """If the member is muted in VC (output)."""

    permissions: Optional[int]
    """Total permissions of the member in the channel, including overwrites, 
        returned when in the interaction object. [`INT_LIMIT`]
    """



from dataclasses import dataclass
from typing import Any

from ..core.http import HTTPClient

@dataclass
class BaseResource:
    """Represents a Discord Resource object."""

    _http: HTTPClient
    """HTTP session for requests."""

    context: Any
    """Associated user data."""



from dataclasses import dataclass
from ..core.model import DataModel

from urllib.parse import quote

@dataclass
class EmojiModel(DataModel):
    """Represents a Discord emoji."""
    
    name: str
    """Name of emoji."""

    id: int = 0
    """ID of the emoji (if custom)."""

    animated: bool = False
    """If the emoji is animated. Defaults to False."""

    @property
    def mention(self) -> str:
        """For use in message content."""
        return f"<a:{self.name}:{self.id}>" if self.animated else f"<:{self.name}:{self.id}>"

    @property
    def api_code(self) -> str:
        """Return the correct API code for this emoji (URL-safe)."""
        if not self.id:
            # unicode emoji
            return quote(self.name)

        # custom emoji
        if self.animated:
            return quote(f"a:{self.name}:{self.id}")
        
        return quote(f"{self.name}:{self.id}")

    @property
    def url(self) -> str:
        """
            Return the full qualifying link for this emoji.

            !!! warning "Important"
                This only works for custom Discord emojis (those with an ID). 
                Unicode emojis will return `None`.
        """
        if not self.id:
            return None
        
        ext = 'gif' if self.animated else 'png'

        return f"https://cdn.discordapp.com/emojis/{self.id}.{ext}"



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

from ..models.user import UserModel
from ..models.guild import ReadyGuildModel
from ..models.application import ApplicationModel

@dataclass
class ReadyEvent(Event, DataModel):
    """Received when bot goes online."""

    v: int
    """API version number."""

    user: UserModel
    """Information about the user."""

    guilds: list[ReadyGuildModel]
    """List of guilds bot is in."""

    session_id: str
    """Used for resuming connections."""

    resume_gateway_url: str
    """Gateway URL for resuming connections."""

    shard: list[int]
    """Shard information associated with this session."""

    application: ApplicationModel
    """Partial application object. Contains ID and flags."""



from dataclasses import dataclass
from ..core.model import DataModel

from .image_data import ImageData

@dataclass
class CreateBotEmoji(DataModel):
    """Represents fields for creating a bot emoji."""
    
    name: str = None
    """Name of the emoji."""
    
    image: ImageData = None
    """Image data for the icon of the emoji."""

@dataclass
class EditBotEmoji(DataModel):
    """Represents fields for editing a bot's emoji."""

    name: str = None
    """Name of the emoji."""
    


from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

from .user import UserModel

@dataclass
class MessageModel(DataModel):
    """A Discord message."""

    id: int
    """ID of the message."""

    channel_id: int
    """Channel ID of the message."""

    author: UserModel
    """User data of author of the message."""
    
    content: str
    """Content of the message."""

    pinned: bool
    """If the message is pinned."""

    type: int
    """Type of message."""

    webhook_id: Optional[int]
    """ID of the webhook if the message is a webhook."""

    timestamp: Optional[str]
    """Timestamp of when the message was sent."""

@dataclass
class PinnedMessageModel(DataModel):
    """Pinned message data."""

    message: MessageModel
    """Message resource of the pinned message."""

    pinned_at: Optional[str]
    """ISO8601 timestamp of when the message was pinned."""



from dataclasses import dataclass

from .base_resource import BaseResource

from ..parts.channel import CreateGuildChannel
from ..parts.role import CreateGuildRole, EditGuildRole

from ..models.role import RoleModel
from ..models.guild import GuildModel
from ..models.user import GuildMemberModel
from ..models.channel import ChannelModel

@dataclass
class Guild(BaseResource):
    """Represents a Discord guild."""
    
    id: int
    """ID of the guild."""

    async def fetch(self, with_counts: bool = False) -> GuildModel:
        """Fetch the Guild object by the given ID.

        Args:
            with_counts (bool, optional): return the approximate member and presence counts for the guild. Defaults to `False`.
            
        Returns:
            (GuildModel): queried guild
        """
        params = {'with_counts': with_counts}

        data = await self._http.request('GET', f'/guilds/{self.id}', params=params)

        return GuildModel.from_dict(data)

    async def fetch_channels(self) -> list[ChannelModel]:
        """Fetch this guild's channels.

        Returns:
            (list[ChannelModel]): queried list of the guild's channels
        """
        data = await self._http.request('GET', f'guilds/{self.id}/channels')

        return [ChannelModel.from_dict(channel) for channel in data]

    async def create_channel(self, params: CreateGuildChannel) -> ChannelModel:
        """Create a channel in this guild.

        Permissions:
            * `MANAGE_CHANNELS` → required to create a channel

        Args:
            params (CreateGuildChannel): the guild channel to create

        Returns:
            (ChannelModel): created channel
        """
        data = await self._http.request('POST', f'/guilds/{self.id}/channels', data=params.to_dict())

        return ChannelModel.from_dict(data)

    async def fetch_guild_member(self, user_id: int) -> GuildMemberModel:
        """Fetch a member in this guild.

        !!! warning "Important"
            Requires the `GUILD_MEMBERS` privileged intent!

        Args:
            user_id (int): user ID of the member to fetch

        Returns:
            (GuildMemberModel): queried guild member
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/members/{user_id}')

        return GuildMemberModel.from_dict(data)
    
    async def fetch_guild_members(self, limit: int = 1, after: int = None) -> list[GuildMemberModel]:
        """Fetch guild members in this guild.

        !!! warning "Important"
            Requires the `GUILD_MEMBERS` privileged intent!

        Args:
            limit (int, optional): Max number of members to return Range 1 - 1000. Default `1`.
            after (int, optional): highest user ID in previous page

        Returns:
            (list[GuildMemberModel]): queried list of guild members
        """
        params = {
            "limit": limit, 
            "after": after
        }

        data = await self._http.request('GET', f'/guilds/{self.id}/members', params=params)

        return [GuildMemberModel.from_dict(member) for member in data]

    async def add_guild_member_role(self, user_id: int, role_id: int) -> None:
        """Append a role to a guild member of this guild.

        Permissions:
            * `MANAGE_ROLES` → required to add a role to the user
        
        Args:
            user_id (int): ID of the member for the role
            role_id (int): ID of the role to append
        """
        await self._http.request('PUT', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')
    
    async def remove_guild_member_role(self, user_id: int, role_id: int) -> None:
        """Remove a role from a guild member of this guild.

        Permissions:
            * `MANAGE_ROLES` → required to remove a role from the user

        Args:
            user_id (int): ID of the member with the role
            role_id (int): ID of the role to remove
        """
        await self._http.request('DELETE', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')

    async def fetch_guild_role(self, role_id: int) -> RoleModel:
        """Fetch a role in this guild.

        Args:
            role_id (int): ID of the role to fetch

        Returns:
            (RoleModel): queried guild role
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/roles/{role_id}')
        
        return RoleModel.from_dict(data)

    async def fetch_guild_roles(self) -> list[RoleModel]:
        """Fetch all roles in this guild.

        Returns:
            (list[RoleModel]): queried list of guild roles
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/roles')
        
        return [RoleModel.from_dict(role) for role in data]

    async def create_guild_role(self, params: CreateGuildRole) -> RoleModel:
        """Create a role in this guild.

        Permissions:
            * `MANAGE_ROLES` → required to add a role to the guild

        Args:
            params (CreateGuildRole): fields to create a role

        Returns:
            (RoleModel): created role
        """
        data = await self._http.request('POST', f'/guilds/{self.id}/roles', data=params.to_dict())

        return RoleModel.from_dict(data)

    async def edit_guild_role(self, role_id: int, params: EditGuildRole) -> RoleModel:
        """Edit a role in this guild.

        Permissions:
            * `MANAGE_ROLES` → required to edit a role in the guild

        Args:
            params (EditGuildRole): role with fields to edit

        Returns:
            (RoleModel): edited role
        """
        data = await self._http.request('PATCH', f'/guilds/{self.id}/roles/{role_id}', data=params.to_dict())

        return RoleModel.from_dict(data)
    
    async def delete_guild_role(self, role_id: int) -> None:
        """Delete a role in this guild.

        Permissions:
            * `MANAGE_ROLES` → required to delete a role in the guild

        Args:
            role_id (int): ID of role to delete
        """
        await self._http.request('DELETE', f'/guilds/{self.id}/roles/{role_id}')



# scurrypy

from .client import Client

__all__ = [
    # top-level modules
    "Client"
]

# imports listed  __all__ libs
from .events import *
from .parts import *
from .resources import *
from .models import *
from .core import *



class DiscordError(Exception):
    """Represents a Discord API error."""

    def __init__(self, status: int, data: dict):
        """Initialize the error with Discord's response.
            Extracts reason, code, and walks the nested errors.

        Args:
            data (dict): Discord's error JSON
        """
        self.data = data
        self.status = status
        self.reason = data.get('message', data)
        self.code = data.get('code', 'Unknown Code')

        self.error_data = data.get('errors', {})
        self.details = self.walk(self.error_data)

        self.is_fatal = status in (401, 403)

        errors = [f"→ {path}: {reason}" for path, reason in self.details]
        full_message = f"{self.reason} ({self.code})"
        if errors:
            full_message += '\n' + '\n'.join(errors)

        super().__init__(full_message)

    def walk(self, node: dict, path=None):
        """Recursively traverses errors field to flatten nested validation errors into (path, message).

        Args:
            node (dict): current error level
            path (tuple[str, str], optional): path to this error level

        Returns:
            (list): list of errors
        """
        if path is None:
            path = []
        result = []

        if isinstance(node, dict):
            for key, value in node.items():
                if key == '_errors' and isinstance(value, list):
                    msg = value[0].get('message', 'Unknown error')
                    result.append(('.'.join(path), msg))

                # the value should not be a dict -- keep going
                elif isinstance(value, dict):
                    result.extend(self.walk(value, path + [key]))
        return result



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

from .emoji import EmojiModel
from .role import RoleModel

@dataclass
class ReadyGuildModel(DataModel):
    """Guild info from Ready event."""
    
    id: int
    """ID of the associated guild."""

    unavailable: bool
    """If the guild is offline."""

@dataclass
class UnavailableGuild(DataModel):
    """Guild info during an outage or before bot bootup."""

    id: int
    """ID of the associated guild."""

    unavailable: bool
    """If the guild is offline."""

@dataclass
class GuildModel(DataModel):
    """Represents a Discord guild."""

    id: int
    """ID of the guild."""
    
    name: str
    """Name of the guild."""

    icon: str
    """Image hash of the guild's icon."""

    splash: str
    """Image hash of the guild's splash."""

    owner: Optional[bool]
    """If the member is the owner."""

    owner_id: int
    """OD of the owner of the guild."""

    roles: list[int]
    """List of IDs registered in the guild."""

    emojis: list[EmojiModel]
    """List of emojis registered in the guild."""

    roles: list[RoleModel]
    """Roles in the guild."""

    mfa_level: int
    """Required MFA level of the guild."""

    application_id: int
    """ID of the application if the guild is created by a bot."""

    system_channel_id: int
    """Channel ID where system messages go (e.g., welcome messages, boost events)."""

    system_channel_flags: int
    """System channel flags."""

    rules_channel_id: int
    """Channel ID where rules are posted."""

    max_members: Optional[int]
    """Maximum member capacity for the guild."""

    description: str
    """Description of the guild."""

    banner: str
    """Image hash of the guild's banner."""

    preferred_locale: str
    """Preferred locale of the guild."""

    public_updates_channel_id: int
    """Channel ID of announcement or public updates."""

    approximate_member_count: int
    """Approximate number of members in the guild."""

    nsfw_level: int
    """NSFW level of the guild."""

    safety_alerts_channel_id: int
    """Channel ID for safety alerts."""



from typing import TypedDict, Unpack

class PermissionsFlagParams(TypedDict, total=False):
    """Role/User permission selection parameters.
    !!! important
        Param name MUST match the permission it represents! 
        For example, "KICK_MEMBERS" is "kick_members".
    """
    create_instant_invite: bool
    kick_members: bool
    ban_members: bool
    administrator: bool
    manage_channels: bool
    manage_guild: bool
    add_reactions: bool
    view_channel: bool
    send_messages: bool
    manage_messages: bool
    embed_links: bool
    attach_files: bool
    read_message_history: bool
    use_external_emojis: bool
    manage_roles: bool
    manage_guild_expressions: bool
    use_application_commands: bool
    manage_threads: bool
    create_public_threads: bool
    create_private_threads: bool
    send_messages_in_threads: bool
    moderate_members: bool
    create_guild_expressions: bool
    send_polls: bool
    pin_messages: bool
    bypass_slowmode: bool

class Permissions:
    """Guild/User permission flags.
    
    !!! note
        Not all permissions are listed. Permissions not listed are not yet supported.
    """

    CREATE_INSTANT_INVITE = 1 << 0
    """Allows creation of instant invites."""

    KICK_MEMBERS = 1 << 1
    """Allows kicking members."""

    BAN_MEMBERS = 1 << 2
    """Allows banning members."""

    ADMINISTRATOR = 1 << 3
    """Allows all permissions and bypasses channel permission overwrites."""

    MANAGE_CHANNELS = 1 << 4
    """Allows management and editing of channels"""

    MANAGE_GUILD = 1 << 5
    """Allows management and editing of the guild."""

    ADD_REACTIONS = 1 << 6
    """Allows for adding new reactions to messages. 
        !!! warning 
            This permission does not apply to reacting with an existing reaction on a message.
    """

    VIEW_CHANNEL = 1 << 10
    """Allows guild members to view a channel, which includes reading messages in text channels and joining voice channels."""

    SEND_MESSAGES = 1 << 11
    """Allows for sending messages in a channel and creating threads in a forum. 
        !!! warning
            Does not allow sending messages in threads.
    """

    MANAGE_MESSAGES = 1 << 13
    """Allows for deletion of other users messages."""

    EMBED_LINKS = 1 << 14
    """Links sent by users with this permission will be auto-embedded."""

    ATTACH_FILES = 1 << 15
    """Allows for uploading images and files."""

    READ_MESSAGE_HISTORY = 1 << 16
    """Allows for reading of message history."""

    USE_EXTERNAL_EMOJIS = 1 << 18
    """Allows the usage of custom emojis from other servers."""

    MANAGE_ROLES = 1 << 28
    """Allows management and editing of roles."""

    MANAGE_GUILD_EXPRESSIONS = 1 << 30
    """Allows for editing and deleting emojis, stickers, and soundboard sounds created by all users."""

    USE_APPLICATION_COMMANDS = 1 << 31
    """Allows members to use application commands, including slash commands and context menu commands."""

    MANAGE_THREADS = 1 << 34
    """Allows for deleting and archiving threads, and viewing all private threads."""

    CREATE_PUBLIC_THREADS = 1 << 35
    """Allows for creating public and announcement threads."""

    CREATE_PRIVATE_THREADS = 1 << 36
    """Allows for creating private threads."""

    SEND_MESSAGES_IN_THREADS = 1 << 38
    """Allows for sending messages in threads."""
    
    MODERATE_MEMBERS = 1 << 40
    """Allows for timing out users to prevent them from sending or reacting to messages in chat and threads, and from speaking in voice and stage channels."""

    CREATE_GUILD_EXPRESSIONS = 1 << 43
    """Allows for creating emojis, stickers, and soundboard sounds, and editing and deleting those created by the current user."""

    SEND_POLLS = 1 << 49
    """Allows sending polls."""
    
    PIN_MESSAGES = 1 << 51
    """Allows pinning and unpinning messages."""

    BYPASS_SLOWMODE = 1 << 52
    """Allows bypassing slowmode restrictions."""

    @staticmethod
    def set(**flags: Unpack[PermissionsFlagParams]):
        """Set permissions. See [`Permission`][scurrypy.core.permissions.Permissions].

        Args:
            **flags (Unpack[PermissionsFlagParams]): permissions to set

        Raises:
            (ValueError): invalid flag

        Returns:
            (int): combined permissions field
        """

        perms = 0
        for k, v in flags.items():
            if v:
                try:
                    perms |= getattr(Permissions, k.upper())
                except AttributeError:
                    raise ValueError(f"Unknown intent flag: '{k}'")
        
        return perms

    @staticmethod
    def has(permissions: int, permission: int):
        """Checks if a permission bit is toggled.

        Args:
            permissions (int): permissions integer
            permission (int): permission bit to check

        Returns:
            (bool): if permission bit is toggled
        """
        return (permissions & permission) == permission



# scurrypy/events

from .channel_events import (
    # GuildChannelEvent,
    GuildChannelCreateEvent,
    GuildChannelUpdateEvent,
    GuildChannelDeleteEvent,
    ChannelPinsUpdateEvent,
)

# from .gateway_events import (
#     SessionStartLimit,
#     GatewayEvent
# )

from .guild_events import (
    # GuildEvent,
    GuildCreateEvent,
    GuildUpdateEvent,
    GuildDeleteEvent,

    GuildMemberAddEvent,
    GuildMemberUpdateEvent,
    GuildMemberRemoveEvent,
    GuildEmojisUpdateEvent
)

# from .hello_event import HelloEvent

from .interaction_events import (
    # ResolvedData,
    # ApplicationCommandOptionData,
    # ApplicationCommandData,
    # MessageComponentData,
    # ModalComponentData,
    # ModalComponent,
    # ModalData,
    InteractionEvent
)

from .message_events import (
    MessageCreateEvent,
    MessageUpdateEvent,
    MessageDeleteEvent,
)

from .reaction_events import (
    ReactionType,
    ReactionAddEvent,
    ReactionRemoveEvent,
    ReactionRemoveEmojiEvent,
    ReactionRemoveAllEvent,
)

from .ready_event import ReadyEvent

from .role_events import (
    RoleCreateEvent,
    RoleUpdateEvent,
    RoleDeleteEvent
)

from .base_event import Event

from .event_types import EventTypes

__all__ = [
    "GuildChannelCreateEvent", "GuildChannelUpdateEvent", "GuildChannelDeleteEvent", "ChannelPinsUpdateEvent",
    "GuildCreateEvent", "GuildUpdateEvent", "GuildDeleteEvent",
    "GuildMemberAddEvent", "GuildMemberRemoveEvent", "GuildMemberUpdateEvent", "GuildEmojisUpdateEvent",
    "InteractionEvent",
    "MessageCreateEvent", "MessageUpdateEvent", "MessageDeleteEvent",
    "ReactionType", "ReactionAddEvent", "ReactionRemoveEvent", "ReactionRemoveEmojiEvent", "ReactionRemoveAllEvent",
    "ReadyEvent", 
    "RoleCreateEvent", "RoleUpdateEvent", "RoleDeleteEvent",
    "Event", "EventTypes"
]



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

from ..models.user import UserModel
from ..models.guild import GuildModel

class ApplicationFlags:
    """Application flags (bitwise constants)."""

    GATEWAY_PRESENCE = 1 << 12
    """Privileged intent to receive presence_update events."""

    GATEWAY_PRESENCE_LIMITED = 1 << 13
    """Intent to receive presence_update events."""

    GATEWAY_GUILD_MEMBERS = 1 << 14
    """Privileged intent to receive member-related events."""

    GATEWAY_GUILD_MEMBERS_LIMITED = 1 << 15
    """Intent to receive member-related events."""

    VERIFICATION_PENDING_GUILD_LIMIT = 1 << 16
    """Indicates unusual growth of an app that prevents verification."""

    GATEWAY_MESSAGE_CONTENT = 1 << 18
    """Privileged intent to receive message content."""

    GATEWAY_MESSAGE_CONTENT_LIMITED = 1 << 19
    """Intent to receive message content."""

@dataclass
class ApplicationModel(DataModel):
    """Represents a Discord application."""

    id: int
    """ID of the application."""

    name: str
    """Name of the application."""

    icon: Optional[str]
    """Icon hash of the application."""

    description: Optional[str]
    """Description of the application."""

    bot_public: Optional[bool]
    """If the application is public."""

    bot_require_code_grant: Optional[bool]
    """If full OAuth2 code grant is required."""

    bot: Optional[UserModel]
    """Partial bot user object of the application."""

    terms_of_service_url: Optional[str]
    """Terms of Service URL of the application"""

    privacy_policy: Optional[str]
    """Privacy Policy URL of the application."""

    owner: Optional[UserModel]
    """Partial user object of the owner of the application."""

    guild_id: Optional[int]
    """Guild ID associated with the application."""

    guild: Optional[GuildModel]
    """Partial guild object of the associated guild."""

    cover_image: Optional[str]
    """Image hash of rich presence invite cover."""

    flags: Optional[int]
    """Public flags of the application.
        See [`ApplicationFlags`][scurrypy.models.application.ApplicationFlags].
    """

    approximate_guild_count: Optional[int]
    """Approximate guild count of the guilds that installed the application."""



from ..events import *

from ..events.event_types import EventTypes

EVENTS = {
    # startup events
    EventTypes.READY: ReadyEvent,

    # channel events
    EventTypes.CHANNEL_CREATE: GuildChannelCreateEvent,
    EventTypes.CHANNEL_UPDATE: GuildChannelUpdateEvent,
    EventTypes.CHANNEL_DELETE: GuildChannelDeleteEvent,
    
    EventTypes.CHANNEL_PINS_UPDATE: ChannelPinsUpdateEvent,

    # guild events
    EventTypes.GUILD_CREATE: GuildCreateEvent,
    EventTypes.GUILD_UPDATE: GuildUpdateEvent,
    EventTypes.GUILD_DELETE: GuildDeleteEvent,

    EventTypes.GUILD_MEMBER_ADD: GuildMemberAddEvent,
    EventTypes.GUILD_MEMBER_UPDATE: GuildMemberUpdateEvent,
    EventTypes.GUILD_MEMBER_REMOVE: GuildMemberRemoveEvent,

    EventTypes.GUILD_EMOJIS_UPDATE: GuildEmojisUpdateEvent,

    # interaction events
    EventTypes.INTERACTION_CREATE: InteractionEvent,

    # message events
    EventTypes.MESSAGE_CREATE: MessageCreateEvent,
    EventTypes.MESSAGE_UPDATE: MessageUpdateEvent,
    EventTypes.MESSAGE_DELETE: MessageDeleteEvent,

    # reaction events
    EventTypes.MESSAGE_REACTION_ADD: ReactionAddEvent,
    EventTypes.MESSAGE_REACTION_REMOVE: ReactionRemoveEvent,
    EventTypes.MESSAGE_REACTION_REMOVE_ALL: ReactionRemoveAllEvent,
    EventTypes.MESSAGE_REACTION_REMOVE_EMOJI: ReactionRemoveEmojiEvent,

    EventTypes.ROLE_CREATE: RoleCreateEvent,
    EventTypes.ROLE_UPDATE: RoleUpdateEvent,
    EventTypes.ROLE_DELETE: RoleDeleteEvent
}



from dataclasses import dataclass, field
from ..core.model import DataModel

from typing import Optional

class CommandTypes:
    CHAT_INPUT = 1
    USER_COMMAND = 2
    MESSAGE_COMMAND = 3

class CommandOptionTypes:
    """Slash command option input types."""

    STRING = 3
    """string (text)"""

    INTEGER = 4
    """integer (Any integer between -2^53+1 and 2^53-1)"""

    BOOLEAN = 5
    """boolean (true/false)"""

    USER = 6
    """user pangination"""

    CHANNEL = 7
    """channel pangination (category and channels)"""

    ROLE = 8
    """role pangination"""

    MENTIONABLE = 9
    """any pangination (role and user)"""

    NUMBER = 10
    """number (Any double between -2^53 and 2^53)"""

    ATTACHMENT = 11
    """file upload (See [Attachment][scurrypy.parts.message.Attachment])"""

@dataclass
class CommandOptionChoice(DataModel):
    """Choice for a command option."""

    name: str = None
    """Name of the choice."""

    value: str | int | float = None
    """Value for the user to select (same as option type)."""

    name_localizations: Optional[dict] = None
    """Dictionary with keys in available locales."""

@dataclass
class CommandOption(DataModel):
    """Option for a slash command."""

    type: int = None
    """Type of option. See [`CommandOptionTypes`][scurrypy.parts.command.CommandOptionTypes]."""

    name: str = None
    """Name of option."""

    description: str = None
    """Description of option."""

    required: bool = False
    """Whether this option is required. Defaults to `False`."""

    choices: list[CommandOptionChoice] = field(default_factory=list)
    """Choices for the user to pick from, max 25. Only valid for STRING, INTEGER, NUMBER option types."""

    autocomplete: bool = False
    """Whether autocomplete interactions are enabled for this option. Defaults to `False`."""

@dataclass
class SlashCommand(DataModel):
    """Represents the slash command object."""

    name: str = None
    """Name of the command."""

    description: str = None
    """Description of the command."""

    options: list[CommandOption] = field(default_factory=list)
    """Parameters or options for the command."""

    type: int = field(init=False, default=CommandTypes.CHAT_INPUT)
    """Command type. Always `CommandTypes.CHAT_INPUT` for this class. See [`CommandTypes`][scurrypy.parts.command.CommandTypes]."""

@dataclass
class UserCommand(DataModel):
    """Represents the user command object."""

    name: str = None
    """Name of the command."""

    type: int = field(init=False, default=CommandTypes.USER_COMMAND)
    """Command type. Always `CommandTypes.USER_COMMAND` for this class. See [`CommandTypes`][scurrypy.parts.command.CommandTypes]."""

@dataclass
class MessageCommand(DataModel):
    """Represents the message command object."""
    
    name: str = None
    """Name of the command."""

    type: int = field(init=False, default=CommandTypes.MESSAGE_COMMAND)
    """Command type. Always `CommandTypes.MESSAGE_COMMAND` for this class. See [`CommandTypes`][scurrypy.parts.command.CommandTypes]."""



from dataclasses import dataclass
from ..core.model import DataModel
from ..core.permissions import Permissions

from typing import Optional

from .channel import ChannelModel
from .guild import GuildModel
from .user import GuildMemberModel
from .message import MessageModel

class InteractionDataTypes:
    """Interaction data types constants."""

    SLASH_COMMAND = 1
    """The interaction is a slash command."""

    USER_COMMAND = 2
    """The interaction is attached to a user."""

    MESSAGE_COMMAND = 3
    """The interaction is attached to a message."""

class InteractionTypes:
    """Interaction types constants."""

    APPLICATION_COMMAND = 2
    """Slash command interaction."""

    MESSAGE_COMPONENT = 3
    """Message component interaction (e.g., button, select menu, etc.)."""

    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    """Application command autocompletion."""

    MODAL_SUBMIT = 5
    """Modal submit interaction."""

class InteractionCallbackTypes:
    """Interaction callback types constants."""

    PONG = 1
    """Acknowledge a Ping."""

    CHANNEL_MESSAGE_WITH_SOURCE = 4
    """Respond to an interaction with a message."""

    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    """Acknowledge an interaction and edit a response later. User sees a loading state."""

    DEFERRED_UPDATE_MESSAGE = 6
    """
        Acknowledge an interaction and edit the original message later. 
        The user does NOT see a loading state. (Components only)
    """

    UPDATE_MESSAGE = 7
    """Edit the message in which the component was attached."""

    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    """Respond to an autocomplete interaction with suggested choices."""

    MODAL = 9
    """Respond to an interaction with a popup modal (not available for MODAL_SUBMIT and PING interactions)."""

    LAUNCH_ACTIVITY = 12
    """Launch an activity associated with the app (Activities must be enabled)."""

@dataclass
class InteractionCallbackDataModel(DataModel):
    """Represents the interaction callback object."""

    id: int
    """ID of the interaction."""

    type: int
    """Type of interaction.
        See [`InteractionCallbackTypes`][scurrypy.models.interaction.InteractionCallbackTypes].
    """

    activity_instance_id: str
    """Instance ID of activity if an activity was launched or joined."""

    response_message_id: int
    """ID of the message created by the interaction."""

    response_message_loading: bool
    """If the interaction is in a loading state."""

    response_message_ephemeral: bool
    """If the interaction is ephemeral."""

@dataclass
class InteractionCallbackModel(DataModel):
    """Represents the interaction callback response object."""

    interaction: InteractionCallbackDataModel
    """The interaction object associated with the interaction response."""

@dataclass
class InteractionModel(DataModel):
    """Represents the interaction model."""

    type: int
    """Type of interaction.
        See [`InteractionTypes`][scurrypy.models.interaction.InteractionTypes].
    """

    id: int
    """ID of interaction."""

    token: str
    """token of interaction."""

    channel_id: int
    """ID of the channel where the interaction was sent."""

    application_id: int
    """ID of the application that owns the interaction."""

    app_permissions: int
    """Bitwise set of permissions pertaining to the location of the interaction. [`INT_LIMIT`]"""

    member: GuildMemberModel # guild member invoking the interaction
    """Guild member invoking the interaction."""

    message: Optional[MessageModel]
    """Message associated with interaction (components or modals)."""

    locale: str
    """Invoking user's locale."""

    guild_locale: str
    """Locale of the guild the interaction was invoked (if invoked in a guild)."""

    guild_id: Optional[int]
    """ID of guild the interaction was invoked (if invoked in a guild)."""

    guild: Optional[GuildModel]
    """Partial guild object of the guild the interaction was invoked (if invoked in a guild)."""

    channel: Optional[ChannelModel]
    """Partial channel object the interaction was invoked."""

    def bot_can(self, permission_bit: int):
        """Checks `app_permissions` to see if permission bit is toggled.

        Args:
            permission_bit (int): permission bit. See [Permissions][scurrypy.core.permissions.Permissions].

        Returns:
            (bool): whether the bot has this permission
        """
        return Permissions.has(self.app_permissions, permission_bit)



from dataclasses import dataclass, fields, is_dataclass
from typing import get_args, get_origin, Union

"""Extract the type from Optional[t]."""
unwrap_optional = lambda t: get_args(t)[0] if get_origin(t) is Union else t

@dataclass
class DataModel:    
    """DataModel is a base class for Discord JSONs that provides 
        hydration from raw dicts, and optional field defaults.
    """

    @classmethod
    def from_dict(cls, data: dict):
        """Hydrates the given data into the dataclass.

        Args:
            data (dict): the JSON data

        Returns:
            (cls): hydrated dataclass
        """
        if not data:
            return None
        
        def convert(t, v):
            t = unwrap_optional(t)
            o = get_origin(t)
            
            # missing field
            if v is None:
                return None
            
            if t is bool:
                return v == 'true'
            
            if is_dataclass(t):
                return t.from_dict(v)
            
            if o is dict:
                vt = get_args(t)[1]
                return {
                    int(k): convert(vt, x) 
                    for k, x in v.items()
                }
            
            if o is list:
                lt = get_args(t)[0]
                return [convert(lt, x) for x in v]
            
            # primitive / fallback
            return t(v)
        
        kwargs = {
            f.name: convert(f.type, data.get(f.name)) 
            for f in fields(cls)
        }

        return cls(**kwargs)
        
    def to_dict(self):
        """Recursively turns the dataclass into a dictionary and drops empty fields.

        Returns:
            (dict): serialized dataclasss
        """
        def serialize(val):
            if isinstance(val, list):
                return [serialize(v) for v in val if v is not None]
            if isinstance(val, DataModel):
                return val.to_dict()
            return val

        result = {}
        for f in fields(self):
            if f.name.startswith('_'):
                continue
            val = getattr(self, f.name)
            # if val not in (None, [], {}, "", 0):
            result[f.name] = serialize(val)
        return result



from dataclasses import dataclass

from .base_resource import BaseResource

from ..models.user import GuildMemberModel
from ..models.user import UserModel

@dataclass
class User(BaseResource):
    """A Discord user."""

    id: int
    """ID of the user."""

    async def fetch(self) -> UserModel:
        """Fetch this user by ID.

        Returns:
            (UserModel): queried user
        """
        data = await self._http.request('GET', f'/users/{self.id}')

        return UserModel.from_dict(data)

    async def fetch_guild_member(self, guild_id: int) -> GuildMemberModel:
        """Fetch this user's guild member data.

        Args:
            guild_id (int): ID of guild to fetch data from

        Returns:
            (GuildMemberModel): queried guild member for the user
        """
        data = await self._http.request('GET', f'/guilds/{guild_id}/members/{self.id}')

        return GuildMemberModel.from_dict(data)



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

@dataclass
class RoleColorModel(DataModel):
    """Role color data."""

    primary_color: int
    """Primary color of the role."""

    secondary_color: int
    """Secondary color of the role. Creates a gradient."""

    tertiary_color: int
    """Tertiary color of the role. Creates a holographic style."""

@dataclass
class RoleModel(DataModel):
    """Represents a Discord role."""

    id: int
    """ID of the role."""

    name: str
    """Name of the role."""

    colors: RoleColorModel
    """Colors of the role."""

    hoist: bool
    """If the role is pinned in user listing."""

    position: int
    """Position of the role."""

    permissions: int
    """Permission bit set. [INT_LIMIT]"""

    managed: bool
    """If the role is managed by an integration."""

    mentionable: bool
    """If the role is mentionable."""

    flags: int
    """Role flags combined as a bitfield."""

    icon: Optional[str]
    """Icon hash of the role."""

    permissions: str
    """permission bits set."""

    unicode_emoji: Optional[str]
    """Unicode emoji of the role."""



from dataclasses import dataclass, field
from ..core.model import DataModel

from typing import Optional

@dataclass
class EmbedAuthor(DataModel):
    """Embed author parameters."""

    name: str = None
    """Name of the author."""

    url: Optional[str] = None
    """URL of the author. http or attachment://<filename> scheme."""

    icon_url: Optional[str] = None
    """URL of author's icon. http or attachment://<filename> scheme."""

@dataclass
class EmbedThumbnail(DataModel):
    """Embed thumbnail."""

    url: str = None
    """Thumbnail content. http or attachment://<filename> scheme."""

@dataclass
class EmbedField(DataModel):
    """Embed field."""

    name: str = None
    """Name of the field."""

    value: str = None
    """Value of the field."""

    inline: Optional[bool] = False
    """Whether or not this field should display inline. Defaults to `False`."""

@dataclass
class EmbedImage(DataModel):
    """Embed image."""

    url: str = None
    """Image content. http or attachment://<filename> scheme."""

@dataclass
class EmbedFooter(DataModel):
    """Embed footer."""

    text: str = None
    """Footer text."""

    icon_url: Optional[str] = None
    """URL of the footer icon. http or attachment://<filename> scheme."""

@dataclass
class EmbedPart(DataModel):
    """Represents the Embed portion of a message."""

    title: Optional[str] = None
    """This embed's title."""

    description: Optional[str] = None
    """This embed's description."""

    timestamp: Optional[str] = None
    """Timestamp of when the embed was sent."""

    color: Optional[int] = None
    """Embed's accent color."""

    author: Optional[EmbedAuthor] = None
    """Embed's author."""

    thumbnail: Optional[EmbedThumbnail] = None
    """Embed's thumbnail attachment."""

    image: Optional[EmbedImage] = None
    """Embed's image attachment."""

    fields: Optional[list[EmbedField]] = field(default_factory=list)
    """List of embed's fields."""

    footer: Optional[EmbedFooter] = None
    """Embed's footer."""

    def to_dict(self):
        """
        EXCEPTION to the "models contain no custom methods" rule for two reasons:

        1. `to_dict` already exists on all models via inheritance, so overriding it
        does not break the design model.

        2. `Thumbnail` (Component V2) and `EmbedThumbnail` (Embed-only) are extremely
        easy to confuse. This guard catches the mistake early and provides a clear,
        actionable error instead of allowing Discord to return an obscure 400 error.
        """
        from .components_v2 import Thumbnail as V2Thumbnail

        if isinstance(self.thumbnail, V2Thumbnail):
            raise TypeError(
                "EmbedPart.thumbnail received a ComponentV2 Thumbnail.\n"
                "Use scurrypy.EmbedThumbnail(url) for embed thumbnails."
            )
        
        if isinstance(self.image, V2Thumbnail):
            raise TypeError(
                "EmbedPart.image received a ComponentV2 Thumbnail.\n"
                "Use scurrypy.EmbedImage(url) for embed thumbnails."
            )
        
        return super().to_dict()


from dataclasses import dataclass
from typing import Optional

from .base_resource import BaseResource

from ..models.command import ApplicationCommandModel

from ..parts.command import SlashCommand, UserCommand, MessageCommand

@dataclass
class Command(BaseResource):
    """Represents a Discord command."""

    application_id: int
    """Application ID of the commands."""

    id: int = None
    """ID of the command."""

    guild_id: Optional[int] = None
    "Guild ID of command (if on guild-level)."

    async def fetch(self) -> ApplicationCommandModel:
        """Fetches the command object.

        Raises:
            (ValueError): no ID set

        Returns:
            (ApplicationCommandModel): queried application command
        """
        if not self.id:
            raise ValueError("No command ID to fetch.")
        
        endpoint = (
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands/{self.id}" 
            if self.guild_id 
            else f"applications/{self.application_id}/commands/{self.id}"
        )

        data = await self._http.request('GET', endpoint)

        return ApplicationCommandModel.from_dict(data)
    
    async def fetch_all(self) -> list[ApplicationCommandModel]:
        """Fetches ALL commands.

        Returns:
            (list[ApplicationCommandModel]): queried list of application commands
        """
        endpoint = (
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands" 
            if self.guild_id 
            else f"applications/{self.application_id}/commands"
        )

        data = await self._http.request('GET', endpoint)

        return [ApplicationCommandModel.from_dict(cmd) for cmd in data]

    async def create(self, command: SlashCommand | UserCommand | MessageCommand) -> ApplicationCommandModel:
        """Add command to the client.

        !!! danger
            Creating a command with the same name as an existing command in the same scope will overwrite the old command.

        Args:
            command (SlashCommand | UserCommand | MessageCommand): command to register

        Returns:
            (ApplicationCommandModel): created command
        """
        endpoint = (
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands" 
            if self.guild_id 
            else f"applications/{self.application_id}/commands"
        )

        data = await self._http.request('POST', endpoint, data=command.to_dict())

        return ApplicationCommandModel.from_dict(data)

    async def edit(self, command: SlashCommand | UserCommand | MessageCommand) -> ApplicationCommandModel:
        """Edit a command.

        Args:
            command (SlashCommand | UserCommand | MessageCommand): command fields to edit
        
        Raises:
            (ValueError): no ID set

        Returns:
            (ApplicationCommandModel): updated application command
        """
        if not self.id:
            raise ValueError("No command ID to fetch.")
        
        endpoint = (
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands/{self.id}" 
            if self.guild_id 
            else f"applications/{self.application_id}/commands/{self.id}"
        )

        data = await self._http.request('PATCH', endpoint, data=command.to_dict())

        return ApplicationCommandModel.from_dict(data)

    async def delete(self) -> None:
        """Delete this command.

        Raises:
            (ValueError): No ID set
        """
        if not self.id:
            raise ValueError("No command ID to fetch.")
        
        endpoint = (
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands/{self.id}" 
            if self.guild_id 
            else f"applications/{self.application_id}/commands/{self.id}"
        )

        await self._http.request('DELETE', endpoint)

    async def bulk_overwrite(self, commands: list[SlashCommand | UserCommand | MessageCommand]) -> list[ApplicationCommandModel]:
        """Takes a list of application commands, overwriting the existing global or guild command list for this application. 
        
        !!! warning
            Commands that do not already exist will count toward daily application command create limits.

        !!! danger
            This will overwrite all types of application commands: slash commands, user commands, and message commands.

        Args:
            commands (list[SlashCommand  |  UserCommand  |  MessageCommand]): commands to register

        Returns:
            (list[ApplicationCommandModel]): created application commands
        """

        endpoint = (
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands" 
            if self.guild_id 
            else f"applications/{self.application_id}/commands"
        )

        data = await self._http.request('PUT', endpoint, data=[cmd.to_dict() for cmd in commands])

        return [ApplicationCommandModel.from_dict(cmd) for cmd in data]



import asyncio
import inspect

from .core.intents import Intents
from .core.http import HTTPClient
from .core.gateway import GatewayClient
from .core.error import DiscordError

from .events.gateway_events import GatewayEvent

import logging

logger = logging.getLogger(__name__)

class Client:
    """Main entry point for Discord bots.
        Ties together the moving parts: gateway, HTTP and event dispatching.
    """

    token: str
    """Bot's token."""

    intents: int
    """Bot intents for listening to events."""

    _http: HTTPClient
    """HTTP session for requests."""

    shards: list[GatewayClient]
    """Shards as a list of gateways."""

    events: dict[str: list[callable]]
    """Events for the client to listen to."""

    startup_hooks: list[callable]
    """Handlers to call once before the bot starts."""

    shutdown_hooks: list[callable]
    """Handlers to call once after the bot shuts down."""

    def __init__(self, 
        *,
        token: str,
        intents: int = Intents.DEFAULT
    ):
        """
        Args:
            token (str): the bot's token
            intents (int, optional): gateway intents. Defaults to `Intents.DEFAULT`.
        """
        if not isinstance(intents, int):
            raise ValueError("Intents must be an integer.")
        
        self.token = token
        self.intents = intents
        
        self._http = HTTPClient()

        self.shards: list[GatewayClient] = []

        self.events = {}
        self.startup_hooks = []
        self.shutdown_hooks = []

    def add_event_listener(self, event: str, handler):
        """Helper function to register listener functions.

        Args:
            event (str): name of the event to listen
            handler (callable): listener function
        """
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 1:
            raise TypeError(
                f"Event listener '{handler.__name__}' must accept exactly one parameter (event)."
            )
    
        self.events.setdefault(event, []).append(handler)

    def add_startup_hook(self, handler):
        """Helper function to register startup functions.
            Runs once on startup BEFORE READY event.

        Args:
            handler (callable): startup function
        """
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 0:
            raise TypeError(
                f"Startup hook '{handler.__name__}' must accept no parameters."
            )
        
        self.startup_hooks.append(handler)

    def add_shutdown_hook(self, handler):
        """Helper function to register shutdown functions.
            Runs once on shutdown.

        Args:
            handler (callable): shutdown function
        """
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 0:
            raise TypeError(
                f"Shutdown hook '{handler.__name__}' must accept no parameters."
            )

        self.shutdown_hooks.append(handler)

    def application(self, application_id: int):
        """Creates an interactable application resource.

        Args:
            application_id (int): ID of target application

        Returns:
            (Application): the Application resource
        """
        from .resources.application import Application

        return Application(self._http, id=application_id, context=None)
    
    def bot_emoji(self, application_id: int):
        """Creates an interactable bot emoji resource.

        Args:
            application_id (int): ID of target application

        Returns:
            (BotEmojis): the BotEmoji resource
        """
        from .resources.bot_emoji import BotEmoji

        return BotEmoji(self._http, None, application_id)
    
    def guild_emoji(self, guild_id: int):
        """Creates an interactable emoji resource.

        Args:
            guild_id (int): guild ID of target emojis

        Returns:
            (GuildEmoji): the GuildEmoji resource
        """
        from .resources.guild_emoji import GuildEmoji

        return GuildEmoji(self._http, None, guild_id)

    def guild(self, guild_id: int, *, context = None):
        """Creates an interactable guild resource.

        Args:
            guild_id (int): ID of target guild
            context (Any, optional): associated data 

        Returns:
            (Guild): the Guild resource
        """
        from .resources.guild import Guild

        return Guild(self._http, context, guild_id)

    def channel(self, channel_id: int, *, context = None):
        """Creates an interactable channel resource.

        Args:
            channel_id (int): ID of target channel
            context (Any, optional): associated data

        Returns:
            (Channel): the Channel resource
        """
        from .resources.channel import Channel

        return Channel(self._http, context, channel_id)
    
    def command(self, application_id: int, guild_id: int = None, command_id: int = None, *, context = None):
        """Creates an interactable command resource.

        Args:
            application_id (int): bot's user ID
            guild_id (int, optional): ID of guild if command is in guild scope
            command_id (int, optional): ID of command
            context (Any, optional): associated data

        Returns:
            (Command): the Command resource
        """
        from .resources.commands import Command

        return Command(self._http, context, application_id, command_id, guild_id)

    def message(self, channel_id: int, message_id: int, *, context = None):
        """Creates an interactable message resource.

        Args:
            message_id (int): ID of target message
            channel_id (int): channel ID of target message
            context (Any, optional): associated data

        Returns:
            (Message): the Message resource
        """
        from .resources.message import Message

        return Message(self._http, context, message_id, channel_id)
    
    def interaction(self, id: int, token: str, *, context = None):
        """Creates an interactable interaction resource.

        Args:
            id (int): ID of the interaction
            token (str): interaction token
            context (Any, optional): associated data

        Returns:
            (Interaction): the Interaction resource
        """
        from .resources.interaction import Interaction

        return Interaction(self._http, context, id, token)
    
    def user(self, user_id: int, *, context = None):
        """Creates an interactable user resource.

        Args:
            user_id (int): ID of target user
            context (Any, optional): associated data

        Returns:
            (User): the User resource
        """
        from .resources.user import User

        return User(self._http, context, user_id)

    async def listen_shard(self, shard: GatewayClient):
        """Consume a GatewayClient's event queue.

        Args:
            shard (GatewayClient): gateway to listen on
        """

        from .core.events import EVENTS

        while True:
            try:
                dispatch_type, event_data = await shard.event_queue.get()

                if dispatch_type not in self.events.keys():
                    logger.debug(f"SHARD ID {shard.shard_id} DISPATCH -> {dispatch_type}")
                else:
                    logger.info(f"SHARD ID {shard.shard_id} DISPATCH -> {dispatch_type}")

                event_model = EVENTS.get(dispatch_type)
                if not event_model:
                    logger.warning(f"Event {dispatch_type} is not implemented.")
                    continue

                obj = event_model.from_dict(event_data)
                obj.name = dispatch_type
                obj.raw = event_data

                handlers = self.events.get(dispatch_type, [])
                for handler in handlers:
                    try:
                        result = handler(obj)
                        if inspect.isawaitable(result):
                            await result
                    except DiscordError as e:
                        logger.error(e)
                        continue

            except Exception:
                # catastrophic errors (network, shard death, unexpected OP code)
                logger.exception(f"SHARD ID {shard.shard_id}: Dispatcher error")
                continue

    async def _start_shards(self, gateway: GatewayEvent):
        """Starts all shards batching by max_concurrency."""

        # pull important values for easier access
        total_shards = gateway.shards
        batch_size = gateway.session_start_limit.max_concurrency

        tasks = []
        
        for batch_start in range(0, total_shards, batch_size):
            batch_end = min(batch_start + batch_size, total_shards)

            logger.debug(f"Starting shards {batch_start}-{batch_end} of {total_shards}")

            for shard_id in range(batch_start, batch_end):
                shard = GatewayClient(gateway.url, shard_id, total_shards)
                self.shards.append(shard)

                # fire and forget
                tasks.append(asyncio.create_task(shard.start(self.token, self.intents)))
                tasks.append(asyncio.create_task(self.listen_shard(shard)))

            # wait before next batch to respect identify rate limit
            await asyncio.sleep(5)

        return tasks
    
    async def start(self):
        """Starts the HTTP/Websocket client, run startup logic, and registers commands."""
        
        try:
            await self._http.start(self.token)

            data = await self._http.request('GET', '/gateway/bot')

            if not data:
                return

            gateway = GatewayEvent.from_dict(data)

            for hook in self.startup_hooks:
                try:
                    result = hook()
                    if inspect.isawaitable(result):
                        await result
                except Exception:
                    logger.exception("Error in shartup hook")

            tasks = await asyncio.create_task(self._start_shards(gateway))

            # end all ongoing tasks
            await asyncio.gather(*tasks)
            
        except asyncio.CancelledError:
            logger.info("Connection cancelled via KeyboardInterrupt.")
        except Exception:
            logger.error(f"Unhandled client start exception.")
        finally:
            await self._close()

    async def _close(self):
        """Gracefully close HTTP session, websocket connections, and run shutdown logic."""  

        for hook in self.shutdown_hooks:
            try:
                result = hook()
                if inspect.isawaitable(result):
                    await result
            except Exception:
                logger.exception("Error in shutdown hook")
                
        logger.info("Closing HTTP session...")
        await self._http.close()

        # close each connection or shard
        await asyncio.gather(*[shard.close_ws() for shard in self.shards])
    
    def run(self):
        """User-facing entry point for starting the client."""  

        try:
            asyncio.run(self.start())
        except Exception as e:
            logger.error(f"{type(e).__name__} {e}")
        finally:
            logger.info("Bot shutting down.")



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

from .image_data import ImageData

@dataclass
class RoleColors(DataModel):
    """Parameters for setting role colors."""

    primary_color: int = None
    """Primary color of the role."""

    secondary_color: int = None
    """Secondary color of the role. Creates a gradient."""

    tertiary_color: int = None
    """Tertiary color of the role. Creates a holographic style."""

@dataclass
class CreateGuildRole(DataModel):
    """Parameters for creating a role."""

    colors: RoleColors = None
    """Colors of the role."""

    name: str = None
    """Name of the role."""

    permissions: int = 0
    """Permission bit set."""

    hoist: bool = None
    """If the role is pinned in the user listing."""

    mentionable: bool = None
    """If the role is mentionable."""

    unicode_emoji: Optional[str] = None
    """Unicode emoji of the role."""

@dataclass
class EditGuildRole(DataModel):
    """Parameters for editing a role."""

    name: Optional[str] = None
    """Name of the role."""

    permissions: Optional[int] = 0
    """Permission bit set."""

    colors: Optional[RoleColors] = None
    """Colors of the role."""

    hoist: Optional[bool] = None
    """If the role is pinned in the user listing."""

    icon: Optional[ImageData] = None
    """Role's icon image (if guild has `ROLE_ICONS` feature)."""

    unicode_emoji: Optional[str] = None
    """Unicode emoji of the role."""

    mentionable: bool = None
    """If the role is mentionable."""



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

class ApplicationCommandTypes:

    CHAT_INPUT = 1
    """Slash commands; a text-based command that shows up when a user types `/`."""

    USER = 2
    """A UI-based command that shows up when you right click or tap on a user."""
    
    MESSAGE = 3
    """A UI-based command that shows up when you right click or tap on a message."""

    PRIMARY_ENTRY_POINT = 4
    """A UI-based command that represents the primary way to invoke an app's Activity."""

class ApplicationCommandOptionTypes:
    
    STRING = 3
    """string (text)"""

    INTEGER = 4
    """integer (Any integer between -2^53+1 and 2^53-1)"""

    BOOLEAN = 5
    """boolean (true/false)"""

    USER = 6
    """user pangination"""

    CHANNEL = 7
    """channel pangination (category and channels)"""

    ROLE = 8
    """role pangination"""

    MENTIONABLE = 9
    """any pangination (role and user)"""

    NUMBER = 10
    """number (Any double between -2^53 and 2^53)"""

    ATTACHMENT = 11
    """file upload (See [Attachment][scurrypy.parts.message.Attachment])"""

@dataclass
class ApplicationCommandOptionChoiceModel(DataModel):
    """Represents the application command option choice object."""

    name: str
    """Name of the choice."""

    value: str
    """Value for the choice.
    
    !!! note
        Convert based on expected type (str, int or double)
    """

@dataclass
class ApplicationCommandOptionModel(DataModel):
    """Represents the application command option object."""

    type: int
    """Type of command option. See [`ApplicationCommandOptionTypes`][scurrypy.models.command.ApplicationCommandOptionTypes]."""

    name: str
    """Name of the command option."""

    descripton: str
    """Description for the command option."""
    
    required: Optional[bool]
    """Whether this option is required. Defaults to `False`."""

    choices: Optional[list[ApplicationCommandOptionChoiceModel]]
    """Choices for the user to pick from."""

    channel_types: Optional[list[int]]
    """Channels shown will be restricted to these types."""

    min_value: Optional[int]
    """Minimum value allowed."""

    max_value: Optional[int]
    """Maximum value allowed."""

    min_length: Optional[int]
    """Minimum length allowed."""

    max_length: Optional[int]
    """Maximum length allowed."""

    autocomplete: Optional[bool]
    """Whether autocomplete interactions are enabled for this option."""


@dataclass
class ApplicationCommandModel(DataModel):
    """Represents the application command object."""

    id: int
    """Unique ID of command."""

    type: Optional[int]
    """Type of command, defaults to `1`. See [`ApplicationCommandTypes`][scurrypy.models.command.ApplicationCommandTypes]."""

    application_id: int
    """ID of the parent application."""

    guild_id: Optional[int]
    """Guild ID of the command, if not global."""

    name: str
    """Name of the command."""

    description: str
    """Description for `CHAT_INPUT` commands. 
    
    !!! note
        Empty for `USER` and `MESSAGE` commands.
    """

    options: Optional[list[ApplicationCommandOptionModel]]
    """Parameters for the command."""

    default_member_permissions: int
    """Set of permissions represented as a bit set. See [`Permissions`][scurrypy.core.permissions.Permissions]. [`INT_LIMIT`]"""

    nsfw: Optional[bool]
    """Whether the command is age-restricted. Defaults to `False`."""



from dataclasses import dataclass

from .base_resource import BaseResource

from ..parts.channel import EditGuildChannel, CreateThreadFromMessage, EditDMChannel, EditThreadChannel, CreateThreadWithoutMessage
from ..parts.message import MessagePart

from ..models.message import MessageModel, PinnedMessageModel
from ..models.channel import ChannelModel, ThreadMemberModel, ArchivedThreads


@dataclass
class Channel(BaseResource):
    """Represents a Discord guild channel."""

    id: int
    """ID of the channel."""

    async def fetch(self) -> ChannelModel:
        """Fetch the full channel data from Discord.

        Returns:
            (ChannelModel): queried channel
        """
        data = await self._http.request("GET", f"/channels/{self.id}")

        return ChannelModel.from_dict(data)
    
    async def edit(self, params: EditGuildChannel | EditDMChannel | EditThreadChannel) -> ChannelModel:
        """Edit this channel.

        Permissions:
            * `MANAGE_CHANNELS` → required to edit this channel

        Args:
            params (EditGuildChannel): channel fields to edit

        Returns:
            (ChannelModel): updated channel
        """

        data = self._http.request('PATCH', f'/channels/{self.id}', data=params.to_dict())

        return ChannelModel.from_dict(data)
    
    async def delete(self) -> None:
        """Deletes this channel from the server.

        Permissions:
            * `MANAGE_CHANNELS` → required to delete this channel
            * `MANAGE_THREADS` → required to delete a thread
        """
        await self._http.request("DELETE", f"/channels/{self.id}")

    async def fetch_messages(self, limit: int = 50, before: int = None, after: int = None, around: int = None) -> list[MessageModel]:
        """Fetches this channel's messages.

        Permissions:
            * `VIEW_CHANNEL` → required to access channel messages
            * `READ_MESSAGE_HISTORY` → required for user, otherwise no messages are returned

        Args:
            limit (int, optional): Max number of messages to return. Range 1 - 100. Defaults to `50`.
            before (int, optional): get messages before this message ID
            after (int, optional): get messages after this message ID
            around (int, optional): get messages around this message ID

        Returns:
            (list[MessageModel]): queried list of messages
        """
        params = {
            "limit": limit,
            "before": before,
            "after": after,
            "around": around
        }

        data = await self._http.request('GET', f'/channels/{self.id}/messages', params=params)

        return [MessageModel.from_dict(msg) for msg in data]
    
    async def send(self, message: str | MessagePart) -> MessageModel:
        """Send a message to this channel.

        Permissions:
            * `SEND_MESSAGES` → required to create a message in this channel

        Args:
            message (str | MessagePart): content as a string or MessagePart

        Returns:
            (MessageModel): created message
        """
        if isinstance(message, str):
            message = MessagePart(content=message)

        message = message._prepare()

        data = await self._http.request(
            "POST", 
            f"/channels/{self.id}/messages", 
            data=message._prepare().to_dict(),
            files=[fp.path for fp in message.attachments]
        )

        return MessageModel.from_dict(data)
    
    async def create_thread_from_message(self, message_id: int, params: CreateThreadFromMessage) -> ChannelModel:
        """Create a thread from a message.

        Args:
            message_id (int): ID of message to attach thread
            params (CreateThreadFromMessage): fields to create thread

        Returns:
            (ChannelModel): new thread
        """

        data = await self._http.request('POST', f"channels/{self.id}/messages/{message_id}/threads", data=params.to_dict())

        return ChannelModel.from_dict(data)
    
    async def fetch_pins(self, limit: int = 50, before: str = None) -> list[PinnedMessageModel]:
        """Get this channel's pinned messages.

        !!! note
            * Creates a `PUBLIC_THREAD` when called on a `GUILF_TEXT` channel
            * Creates an `ANNOUNCEMENT_THREAD` when called on a `GUILD_ANNOUNCEMENT` channel
        
        !!! warning
            Does not work on a `GUILD_FORUM` channel!

        Permissions:
            * `VIEW_CHANNEL` → required to access pinned messages
            * `READ_MESSAGE_HISTORY` → required for reading pinned messages

        Args:
            before (str, optional): get pinned messages before this ISO8601 timestamp
            limit (int, optional): Max number of pinned messages to return. Range 1 - 50. Defaults to `50`.
        
        Returns:
            (list[PinnedMessage]): queried list of pinned messages
        """
        # Set default limit if user didn't supply one
        params = {
            "limit": limit,
            "before": before
        }

        data = await self._http.request('GET', f'/channels/{self.id}/pins', params=params)

        return [PinnedMessageModel.from_dict(item) for item in data]

    async def create_thread_without_message(self, params: CreateThreadWithoutMessage) -> ChannelModel:
        """Create a thread not connected to an existing message.

        Args:
            params (CreateThreadWithoutMessage): fields to create the thread

        Returns:
            (ChannelModel): new thread
        """

        data = self._http.request('POST', f'/channels/{self.id}/threads', data=params.to_dict())

        return ChannelModel.from_dict(data)

    async def join_thread(self) -> None:
        """Add the bot to this thread.

        !!! important
            Required the thread NOT be archived.
        """
        await self._http.request('PUT', f'/channels/{self.id}/thread-members/@me')
    
    async def add_thread_member(self, user_id: int) -> None:
        """Add a user to this thread.

        Args:
            user_id (int): ID of the user to add
        """
        await self._http.request('PUT', f'/channels/{self.id}/thread-members/{user_id}')

    async def leave_thread(self) -> None:
        """Remove the bot from a thread.

        !!! important
            Required the thread NOT be archived.
        """
        await self._http.request('DELETE', f'/channels/{self.id}/thread-members/@me')
    
    async def remove_thread_member(self, user_id: int) -> None:
        """Remove a user to this thread.

        Args:
            user_id (int): ID of the user to remove
        """
        await self._http.request('DELETE', f'/channels/{self.id}/thread-members/{user_id}')

    async def fetch_thread_member(self, user_id: int, with_member: bool = False) -> ThreadMemberModel:
        """Fetch a thread emmber of the specified user ID from this thread.

        Args:
            user_id (int): ID of the user to fetch
            with_member (bool, optional): whether to include the member object. Defaults to `False`.
        
        Returns:
            (ThreadMemberModel): queried thread member
        """

        params = { 'with_member': with_member }

        data = await self._http.request('GET', f'/channels/{self.id}/thread-members/{user_id}', params=params)

        return ThreadMemberModel.from_dict(data)
    
    async def fetch_thread_members(self, limit: int = 100, after: int = None, with_member: bool = False) -> list[ThreadMemberModel]:
        """Fetch all members of this thread.

        !!! warning
            Requires the `GUILD_MEMBERS` privileged intent to use!

        Args:
            limit (int, optional): Max number of thread members to return. Range 0 - 100. Defaults to `100`.
            after (int, optional): members after this user ID
            with_member (bool, optional): whether to include the member object. Defaults to `False`.

        Returns:
            (list[ThreadMemberModel]): queried list of thread members
        """

        params = {
            'with_member': with_member,
            'after': after,
            'limit': limit
        }

        data = await self._http.request('GET', f"/channels/{self.id}/thread-members", params=params)

        return [ThreadMemberModel.from_dict(n) for n in data]

    async def fetch_public_archived_threads(self, before: str = None, limit: int = None) -> ArchivedThreads:
        """Return public archived threads in this channel.

        !!! note
            Threads are ordered by `active_timestamp` in descending order.

        Permissions:
            * `READ_MESSAGE_HISTORY` → required for reading threads

        Args:
            before (str, optional): threads before this timestamp
            limit (int, optional): max number of threads to return

        Returns:
            (ArchivedThreads): queried archived threads
        """

        params = {
            'before': before,
            'limit': limit
        }

        data = await self._http.request('GET', f'/channels/{self.id}/threads/archived/public', params=params)

        return ArchivedThreads.from_dict(data)

    async def fetch_private_archived_threads(self, before: str = None, limit: int = None) -> ArchivedThreads:
        """Return private archived threads in this channel.

        !!! note
            Threads are ordered by `active_timestamp` in descending order.

        Permissions:
            * `READ_MESSAGE_HISTORY` → required for reading threads
            * `MANAGE_THREADS` → required for accessing threads

        Args:
            before (str, optional): threads before this timestamp
            limit (int, optional): max number of threads to return

        Returns:
            (ArchivedThreads): queried archived threads
        """

        params = {
            'before': before,
            'limit': limit
        }

        data = await self._http.request('GET', f'/channels/{self.id}/threads/archived/private', params=params)

        return ArchivedThreads.from_dict(data)

    async def fetch_joined_private_archived_threads(self, before: int = None, limit: int = None) -> ArchivedThreads:
        """Return private archived threads in this channel that the bot has joined.

        !!! note
            Threads are ordered by their ID in descending order.

        Permissions:
            * `READ_MESSAGE_HISTORY` → required for reading threads

        Args:
            before (int, optional): threads before this ID
            limit (int, optional): max number of threads to return

        Returns:
            (ArchivedThreads): queried archived threads
        """

        params = {
            'before': before,
            'limit': limit
        }

        data = await self._http.request('GET', f'/channels/{self.id}/users/@me/threads/archived/private', params=params)

        return ArchivedThreads.from_dict(data)



# scurrypy/resources

from .application import Application
from .bot_emoji import BotEmoji

from .channel import (
    # MessagesFetchParams,
    # PinsFetchParams,
    # ThreadFromMessageParams,
    Channel
)
from .commands import Command
from .guild_emoji import GuildEmoji

from .guild import (
    # FetchGuildMembersParams,
    # FetchGuildParams,
    Guild
)

from .interaction import Interaction

from .message import Message

from .user import (
    # FetchUserGuildsParams,
    User
)

__all__ = [
    "Application",
    "BotEmoji",
    "Channel",
    "Command",
    "Guild",
    "GuildEmoji",
    "Interaction",
    "Message",
    "User"
]



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

@dataclass
class SessionStartLimit(Event, DataModel):
    """Represents the Session Start Limit object."""

    total: int
    """Total remaining shards."""

    remaining: int
    """Shards left to connect."""

    reset_after: int
    """When `remaining` resets from now (in ms)."""

    max_concurrency: int
    """How many shards can be started at once."""

@dataclass
class GatewayEvent(Event, DataModel):
    """Represents the Gateway Event object."""

    url: str 
    """Gateway URL to connect."""

    shards: int
    """Recommended shard count for the aaplication."""

    session_start_limit: SessionStartLimit
    """Session start info."""



from dataclasses import dataclass

from ..models.application import ApplicationModel

from .base_resource import BaseResource

@dataclass
class Application(BaseResource):
    """Represents a Discord application."""

    id: int
    """ID of the application."""

    async def fetch(self) -> ApplicationModel:
        """Fetch this application's data.

        Returns:
            (Application): queried application
        """
        data = await self._http.request('GET', '/applications/@me')

        return ApplicationModel.from_dict(data)



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

from typing import Optional

from ..models.channel import ChannelModel

class GuildChannelCreateEvent(Event, ChannelModel):
    """Received when a guild channel has been created."""
    pass

class GuildChannelUpdateEvent(Event, ChannelModel):
    """Received when a guild channel has been updated."""
    pass

class GuildChannelDeleteEvent(Event, ChannelModel):
    """Received when a guild channel has been deleted."""
    pass

@dataclass
class ChannelPinsUpdateEvent(Event, DataModel):
    """Pin update event."""
    
    channel_id: int
    """ID of channel where the pins were updated."""

    guild_id: Optional[int]
    """ID of the guild where the pins were updated."""

    last_pin_timestamp: Optional[str]  # ISO8601 timestamp of last pinned message
    """ISO8601 formatted timestamp of the last pinned message in the channel."""



# scurrypy/parts

from .bot_emoji import CreateBotEmoji, EditBotEmoji

from .channel import (
    ChannelTypes,
    ChannelFlags,
    SortOrderTypes,
    ForumLayoutTypes,
    TagPart,
    DefaultReactionPart,
    CreateGuildChannel,
    EditDMChannel,
    EditThreadChannel,
    EditGuildChannel,
    CreateThreadWithoutMessage
)

from .command import (
    CommandTypes,
    CommandOptionTypes,
    CommandOption,
    CommandOptionChoice,
    SlashCommand, 
    UserCommand,
    MessageCommand
)

from .component_types import (
    ContainerChild,
    ActionRowChild,
    LabelChild,
    SectionAccessoryChild,
    SectionChild
)

from .components_v2 import (
    SectionPart,
    TextDisplay,
    Thumbnail,
    MediaGalleryItem,
    MediaGallery,
    File,
    SeparatorTypes,
    Separator,
    ContainerPart,
    Label,
    FileUpload
)

from .components import (
    ComponentTypes,
    ActionRowPart, 
    ButtonStyles,
    Button,
    SelectOption,
    StringSelect,
    TextInputStyles,
    TextInput,
    DefaultValue,
    # SelectMenu,
    UserSelect,
    RoleSelect,
    MentionableSelect,
    ChannelSelect
)

from .embed import (
    EmbedAuthor,
    EmbedThumbnail,
    EmbedField,
    EmbedImage,
    EmbedFooter,
    EmbedPart
)

from .guild_emoji import CreateGuildEmoji, EditGuildEmoji

from .image_data import ImageData

from .message import (
    MessageFlags,
    # MessageFlagParams,
    MessageReferenceTypes,
    MessageReference,
    Attachment,
    MessagePart
)

from .modal import ModalPart
from .role import CreateGuildRole, EditGuildRole, RoleColors

__all__ = [
    "CreateBotEmoji", "EditBotEmoji",
    "CreateGuildChannel", "TagPart", "DefaultReactionPart", "ChannelTypes", "ChannelFlags", "SortOrderTypes", "ForumLayoutTypes", 
    "EditDMChannel", "EditThreadChannel", "EditGuildChannel", "CreateThreadWithoutMessage",
    "CommandTypes", "CommandOption", "CommandOptionChoice", "CommandOptionTypes", "SlashCommand", "UserCommand", "MessageCommand",
    "ContainerChild", "ActionRowChild", "LabelChild", "SectionAccessoryChild", "SectionChild",
    "SectionPart", "TextDisplay", "Thumbnail", "MediaGalleryItem", "MediaGallery",
    "File", "SeparatorTypes", "Separator", "ContainerPart", "Label", "FileUpload",
    "ComponentTypes", "ActionRowPart", "ButtonStyles", "Button", "SelectOption", "StringSelect",
    "TextInputStyles", "TextInput", "DefaultValue", "UserSelect", "RoleSelect", "MentionableSelect",
    "ChannelSelect",
    "EmbedAuthor", "EmbedThumbnail", "EmbedField", "EmbedImage", "EmbedFooter", "EmbedPart",
    "CreateGuildEmoji", "EditGuildEmoji",
    "ImageData",
    "MessageFlags", "MessageReferenceTypes", "MessageReference", "Attachment", "MessagePart", 
    "ModalPart",
    "CreateGuildRole", "EditGuildRole", "RoleColors"
]



from dataclasses import dataclass

from .base_resource import BaseResource

from ..models.emoji import EmojiModel

from ..parts.bot_emoji import *

@dataclass
class BotEmoji(BaseResource):
    """Represents a Discord Bot Emoji."""

    application_id: int
    """Application ID of the emojis."""

    async def fetch(self, emoji_id: int) -> EmojiModel:
        """Fetch an emoji from the bot repository.

        Args:
            emoji_id (int): emoji ID

        Returns:
            (EmojiModel): queried emoji
        """
        data = await self._http.request("GET", f"/applications/{self.application_id}/emojis/{emoji_id}")

        return EmojiModel.from_dict(data)
    
    async def fetch_all(self) -> list[EmojiModel]:
        """Fetch all emojis from the bot repository.

        Returns:
            (list[EmojiModel]): queried list of bot emojis
        """
        data = await self._http.request("GET", f"/applications/{self.application_id}/emojis")

        emojis = data.get("items")

        return [EmojiModel.from_dict(emoji) for emoji in emojis]
    
    async def create(self, params: CreateBotEmoji) -> EmojiModel:
        """Add an emoji to the bot emoji repository.

        Args:
            params (CreateBotEmoji): fields for creating a bot emoji

        Returns:
            (EmojiModel): new emoji
        """
        data = await self._http.request(
            'POST', 
            f'/applications/{self.application_id}/emojis',
            data=params.to_dict()
        )
    
        return EmojiModel.from_dict(data)
    
    async def edit(self, emoji_id: int, params: EditBotEmoji) -> EmojiModel:
        """Edit an emoji in the bot repository.

        Args:
            params (EditBotEmoji): fields for editing a bot emoji

        Returns:
            (EmojiModel): updated emoji
        """
        data = await self._http.request(
            'PATCH', 
            f'/applications/{self.application_id}/emojis/{emoji_id}', 
            data=params.to_dict()
        )

        return EmojiModel.from_dict(data)

    async def delete(self, emoji_id: int) -> None:
        """Deletes an emoji from the bot repository.

        Args:
            emoji_id (int): ID of the emoji to remove
        """
        await self._http.request('DELETE', f'/applications/{self.application_id}/emojis/{emoji_id}')



from dataclasses import dataclass, field
from ..core.model import DataModel

from typing import Optional

from .image_data import ImageData

@dataclass
class CreateGuildEmoji(DataModel):
    """Represents fields for creating a guild emoji."""
    
    name: str = None
    """Name of the emoji."""
    
    image: ImageData = None
    """Image data for the icon of the emoji."""
    
    roles: list[int] = field(default_factory=list)
    """Roles able to use the emoji."""

@dataclass
class EditGuildEmoji(DataModel):
    """Represents fields for editing a guild emoji."""

    name: Optional[str] = None
    """Name of the emoji."""

    roles: Optional[list[int]] = field(default_factory=list)
    """Roles able to use the emoji."""



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional, Literal

from .image_data import ImageData

class ChannelTypes:
    """Represents the types of channels."""

    GUILD_TEXT = 0
    """Text channel within a guild."""

    DM = 1
    """Direct message between users."""

    GROUP_DM = 3
    """Direct message between many users."""

    GUILD_CATEGORY = 4
    """Organizational category for channels."""

    GUILD_ANNOUNCEMENT = 5
    """Channel users can follow and crosspost into their own server."""

    ANNOUNCEMENT_THREAD = 10
    """Temporary sub-channel within a `GUILD_ANNOUNCEMENT` channel."""

    PUBLIC_THREAD = 11
    """Temporary sub-channel within a `GUILD_TEXT` or `GUILD_FORUM` channel."""

    PRIVATE_THREAD = 12
    """Temporary sub-channel within a `GUILD_TEXT` channel only viewable by invitees and members with `MANAGE_THREADS`."""

    GUILD_DIRECTORY = 14
    """Channel in a hub containing the listed servers."""

    GUILD_FORUM = 15
    """Channel that can only contain threads."""

class ChannelFlags:
    """Represents constant bit fields for channel flags."""

    PINNED = 1 << 1
    """This thread is pinned to the top of its parent `GUILD_FORUM` channel."""

    REQUIRE_TAG = 1 << 4
    """Whether a tag is required when creating a thread in a `GUILD_FORUM` channel."""

class SortOrderTypes:
    """Represents sort order types for `GUILD_FORUM` channels."""

    LATEST_ACTIVITY = 0
    """Sort by activity."""

    CREATION_DATE = 1
    """Sort by creation time (recent to oldest)."""

class ForumLayoutTypes:
    """Represents `GUILD_FORUM` layout types."""

    NOT_SET = 0
    """No default layout has been set."""

    LIST_VIEW = 1
    """Display posts as a list."""

    GALLERY_VIEW = 2
    """Display posts as a collection of tiles."""

@dataclass
class DefaultReactionPart(DataModel):
    """Represents the default reaction for a `GUILD_FORUM` post."""

    emoji_id: int = None
    """ID of the guild's custom emoji."""

    emoji_name: str = None
    """Unicode character of the emoji."""

@dataclass
class TagPart(DataModel):
    """Represents the tag object found in `GUILD_FORUM` channels."""
    
    name: str = None
    """Name of the tag."""

    moderated: bool = None
    """Whether the tag can only be added/removed by a member with `MANAGE_THREADS`."""
    
    emoji_id: int = None
    """ID of a guild's custom emoji."""
    
    emoji_name: str = None
    """Unicode character of the emoji."""

@dataclass
class CreateGuildChannel(DataModel):
    """Parameters for creating a guild channel."""

    name: str = None
    """Name of the channel."""

    type: Optional[int] = None
    """Type of channel. See [`ChannelTypes`][scurrypy.parts.channel.ChannelTypes]."""

    position: Optional[int] = None
    """Sorting position of the channel (channels with the same position are sorted by id)."""

    topic: Optional[str] = None
    """Topic of the channel."""

    rate_limit_per_user: Optional[int] = None
    """Seconds user must wait between sending messages in the channel."""

    parent_id: Optional[int] = None
    """Category ID of the channel."""

    nsfw: Optional[bool] = None
    """If the channel is flagged NSFW."""

    default_auto_archive_duration: Optional[int] = None
    """Default duration in minutes threads will be hidden after period of inactivity."""

    default_reaction_emoji: Optional[DefaultReactionPart] = None
    """Emoji to show in the add reaction button in a `GUILD_FORUM` post."""

    available_tags: Optional[list[TagPart]] = None
    """Set of tags that can be applied to a `GUILD_FORUM` post."""

    default_sort_order: Optional[int] = None
    """Default forum sort order. See [`SortOrderTypes`][scurrypy.parts.channel.SortOrderTypes]. Defaults to `None`."""

    default_forum_layout: Optional[int] = None
    """Default forum layout view. See [`ForumLayoutTypes`][scurrypy.parts.channel.ForumLayoutTypes]. Defaults to `ForumLayoutTypes.NOT_SET`."""

    default_thread_rate_limit_per_user: Optional[int] = None
    """Rate limit per user set on newly created threads.
    
    !!! note
        This field does not live update!
    """

@dataclass
class EditGuildChannel(DataModel):
    """Parameters for editing a guild channel."""

    name: str = None
    """Name of the channel."""

    type: Optional[int] = None
    """Type of channel. See [`ChannelTypes`][scurrypy.parts.channel.ChannelTypes]."""

    position: Optional[int] = None
    """Sorting position of the channel (channels with the same position are sorted by id)."""

    topic: Optional[str] = None
    """Topic of the channel."""

    nsfw: Optional[bool] = None
    """If the channel is flagged NSFW."""

    rate_limit_per_user: Optional[int] = None
    """Seconds user must wait between sending messages in the channel."""

    parent_id: Optional[int] = None
    """Category ID of the channel."""

    default_auto_archive_duration: Optional[int] = None
    """Default duration in minutes threads will be hidden after period of inactivity."""

    flags: Optional[int] = None
    """Channel flags. See [`ChannelFlags`][scurrypy.parts.channel.ChannelFlags]."""

    default_reaction_emoji: Optional[DefaultReactionPart] = None
    """Emoji to show in the add reaction button in a `GUILD_FORUM` post."""

    available_tags: Optional[list[TagPart]] = None
    """Set of tags that can be applied to a `GUILD_FORUM` post."""

    default_sort_order: Optional[int] = None
    """Default forum sort order. See [`SortOrderTypes`][scurrypy.parts.channel.SortOrderTypes]. Defaults to `None`."""

    default_forum_layout: Optional[int] = None
    """Default forum layout view. See [`ForumLayoutTypes`][scurrypy.parts.channel.ForumLayoutTypes]. Defaults to `ForumLayoutTypes.NOT_SET`."""

    default_thread_rate_limit_per_user: Optional[int] = None
    """Rate limit per user set on newly created threads.
    
    !!! note
        This field does not live update!
    """

@dataclass
class CreateThreadFromMessage(DataModel):
    """Parameters for creating a thread attached to a message."""

    name: str = None
    """Name of the channel."""

    rate_limit_per_user: Optional[int] = None
    """Seconds user must wait between sending messages in the channel."""

    auto_archive_duration: Optional[Literal[60, 1440, 4320, 10080]] = None
    """Duration in minutes threads will be hidden after period of inactivity."""

@dataclass
class CreateThreadWithoutMessage(DataModel):
    """Parameters for creating a thread without a message."""

    name: str = None
    """Name of the channel."""

    rate_limit_per_user: Optional[int] = None
    """Seconds user must wait between sending messages in the channel."""

    auto_archive_duration: Optional[Literal[60, 1440, 4320, 10080]] = None
    """Duration in minutes threads will be hidden after period of inactivity."""

    type: Optional[int] = ChannelTypes.PRIVATE_THREAD
    """Type of thread to create."""

    invitable: Optional[bool] = None
    """Whether non-moderators can add other non-moderators to the thread (private threads only)."""

@dataclass
class EditDMChannel(DataModel):
    """Parameters for editing a DM channel."""

    name: Optional[str] = None
    """Channel name."""

    icon: Optional[ImageData] = None
    """Channel icon."""

@dataclass
class EditThreadChannel(DataModel):
    """Parameters for editing a thread channel."""

    name: Optional[str] = None
    """Name of the channel."""

    archived: Optional[bool] = None
    """Whether the thread is archived."""

    auto_archive_duration: Optional[Literal[60, 1440, 4320, 10080]] = None
    """Duration in minutes threads will be hidden after period of inactivity."""

    locked: bool = None
    """Whether the thread is locked.
    
    !!! note
        Only users with `MANAGE_THREADS` can unarchive the thread.
    """

    invitable: Optional[bool] = None
    """Whether non-moderators can add other non-moderators to the thread (private threads only)."""

    rate_limit_per_user: Optional[int] = None
    """Seconds user must wait between sending messages in the channel."""

    flags: Optional[int] = None
    """Channel flags. See [`ChannelFlags`][scurrypy.parts.channel.ChannelFlags]."""

    applied_tags: Optional[list[int]]= None
    """Set of tags applied to a `GUILD_FORUM` post."""



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

from typing import Optional

from ..models.message import MessageModel
from ..models.user import GuildMemberModel

@dataclass
class MessageCreateEvent(Event, MessageModel):
    """Received when a message is created. (This event IS the MessageModel with extra fields)
    
    !!! note
        `member` may be missing on `MESSAGE_CREATE` and `MESSAGE_UPDATE`. Use `author` when you need the user.
    """

    guild_id: Optional[int]
    """Guild ID of the updated message (if in a guild channel)."""

    member: Optional[GuildMemberModel]  # guild-only author info
    """Partial Member object of the author of the message. See [`GuildMemberModel`][scurrypy.models.GuildMemberModel]."""

@dataclass
class MessageUpdateEvent(Event, MessageModel):
    """Received when a message is updated. (This event IS the MessageModel with extra fields)"""

    guild_id: Optional[int]
    """Guild ID of the updated message (if in a guild channel)."""

    member: Optional[GuildMemberModel]
    """Partial Member object of the author of the message. See [`GuildMemberModel`][scurrypy.models.GuildMemberModel]."""

@dataclass
class MessageDeleteEvent(Event, DataModel):
    """Received when a message is deleted."""

    id: int
    """ID of the deleted message."""

    channel_id: int
    """Channel ID of the deleted message."""

    guild_id: Optional[int]
    """Guild ID of the deleted message (if in a guild channel)."""
